/**
   @file JavaPlugin.cpp
 
    Implements MPlugInBuilder as defined in Plugin API
 
 	@author Tomaso Paoletti <tpaoletti@users.sourceforge.net>
 
 	Based on Translator sample plugin by John R. Dance, 1999
 	
   This software is distributed under the GNU Public License.  
   See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
*/

#include <stdio.h>
#include <stdlib.h>
#include <algobase.h>

#include <Entry.h>
#include <Path.h>
#include <String.h>

#include <KernelExport.h>

#include "ErrorMessage.h"
#include "MProject.h"

#include "JavaPlugin.h"
#include "JavaPluginPrefs.h"

/** supported compilers **/
#define COMPILER_UNSUPPORTED	0
#define COMPILER_JAVAC			0x1
#define COMPILER_JIKES			0x2
#define COMPILER_GUAVAC			0x4

const char * TOOL_NAME	 = "jikes";			/* name shown in Target panel */	
const char * DEFAULT_TOOL= "jikes";			/* compiler used by default */
const char * DEFAULT_CLASSPATH= "/boot/home/bekaffe/share/kaffe/Klasses.jar";

// ---------------------------------------------------------------------------

int 
JavaPlugin::currentCompiler()
{
	int res = COMPILER_UNSUPPORTED;
	if (strcmp(fPrefs.compiler, "javac")==0 || strcmp(fPrefs.compiler, "pizza")==0)
		res = COMPILER_JAVAC; 
	else if (strcmp(fPrefs.compiler, "jikes")==0) 
		res = COMPILER_JIKES;
	else if (strcmp(fPrefs.compiler, "guavac")==0)
		res = COMPILER_GUAVAC;
	return res;
}

// ---------------------------------------------------------------------------

JavaPlugin::JavaPlugin()
{
	// set default prefs, not necessarily in sync with stored prefs
	strncpy(fPrefs.compiler, DEFAULT_TOOL, MAX_STR); 
	
	fFiles = new BList();
}

// ---------------------------------------------------------------------------

JavaPlugin::~JavaPlugin()
{
	if (fFiles)
		delete fFiles;
}

// ---------------------------------------------------------------------------

status_t 
JavaPlugin::GetToolName(MProject* inProject, char *outName, int32 inBufferLength, MakeStageT inStage, MakeActionT inAction)
{
	strncpy(outName, TOOL_NAME, inBufferLength);
DBG(("ToolName: %s\n", outName));
	return B_OK;
}

// ---------------------------------------------------------------------------

const char *
JavaPlugin::LinkerName()
{
	// we don't deal with linking
	return B_EMPTY_STRING; 
}

// ---------------------------------------------------------------------------

MakeStageT 
JavaPlugin::MakeStages()
{
	// all our work is done during compile
	return kCompileStage; 	
}

// ---------------------------------------------------------------------------

MakeActionT 
JavaPlugin::Actions()
{
	return kCompile;  
}

// ---------------------------------------------------------------------------

PlugInFlagsT 
JavaPlugin::Flags()
{
	return kNotIDEAware;
}

// ---------------------------------------------------------------------------

ulong 
JavaPlugin::MessageDataType()
{
	return kMessageType;
}

// ---------------------------------------------------------------------------


void 
JavaPlugin::ProjectChanged(MProject& inProject, ChangeT inChange)
{	
	switch (inChange)
	{
		case kProjectOpened: case kPrefsChanged:
DBG(("JavaPlugin prefs changed\n"));
			//if (inProject)
			{
			BMessage msg;
				inProject.GetPrefs(kMessageType, msg);
				if (compilerPrefs::FromMessage(msg, fPrefs))
				{
DBG((" comp: %s\n path: %s\n options: %s\n strict: %d\n debug: %d\n nowarn: %d\n",
		fPrefs.compiler, fPrefs.classpath, fPrefs.options, 
		fPrefs.strict, fPrefs.debug, fPrefs.nowarn));
				} else {
DBG(("UNABLE TO LOAD PREFS\n"));
				}				
			}
			break;
			
		case kProjectClosed:
			break;
			
		case kFilesAdded:
DBG(("Files added\n"));
			
			break;
			
		case kFilesRemoved:
		case kFilesRearranged:
		case kRunMenuItemChanged:
		case kLinkDone:
			// Don't do anything for now
			break;
			
		case kBuildStarted:
DBG(("Build started\n"));

			// -- clear list of compiled files
			fFiles->MakeEmpty();
#if 0
			entry_ref ref;
			if (inProject.GetProjectRef(ref)==B_NO_ERROR)
			{	BEntry entry(&ref);
				projectPath.SetTo(&entry);
				projectPath.GetParent(&projectPath);
DBG(("Project path set to %s\n", projectPath.Path() ));
			}
#endif
			break;
	}
}

// ---------------------------------------------------------------------------

bool 
JavaPlugin::ValidateSettings(BMessage &inOutMessage)
{
	return false;
}

// ---------------------------------------------------------------------------

status_t 
JavaPlugin::BuildPrecompileArgv(MProject& inProject, BList &inArgv, MFileRec &inFileRec)
{
/*
	if (inFileRec.makeStage != kPrecompileStage) {
		return B_ERROR;
	}

	return BuildArgv(inArgv, inFileRec.path);
*/
	return B_ERROR;
}

// ---------------------------------------------------------------------------

status_t 
JavaPlugin::BuildCompileArgv(MProject& inProject, BList &inArgv, MakeActionT inAction, MFileRec &inFileRec)
{
	if (inFileRec.makeStage != kCompileStage) {
		return B_ERROR;
	}

	char * filePath = inFileRec.path;

	DBG(("BuildCompileArgv: compile %s (%s)\n", inFileRec.name, inFileRec.path));
	
	// -- add the file name to the list of compiled files (we need this for correct error parsing)
	if (! fFiles->HasItem(&inFileRec)) {
		fFiles->AddItem(&inFileRec);
	}
		
	//BString outputName;
	//this->makeOutputFileName(filePath, outputName);

	// -- compiler argument list
	if (strlen(fPrefs.classpath) > 0)
	{	
		inArgv.AddItem(strdup("-classpath"));
		inArgv.AddItem(strdup(fPrefs.classpath));
	}
	if (fPrefs.debug)
		inArgv.AddItem(strdup("-debug"));
	if (fPrefs.strict)
		inArgv.AddItem(strdup("-java"));
	if (fPrefs.nowarn)
		inArgv.AddItem(strdup("-nowarn"));
	
	if (currentCompiler() == COMPILER_JIKES) {
		// option flag for Emacs-style errors
		inArgv.AddItem(strdup("+E"));
	}
	
	// -- file to compile	
	inArgv.AddItem(strdup(filePath));
	
	return B_OK;
}

// ---------------------------------------------------------------------------

status_t 
JavaPlugin::BuildPostLinkArgv(MProject& inProject, BList &inArgv, MFileRec &inFileRec)
{
	// we don't do anything in this stage
	printf("post link\n");
	return B_ERROR;
}

// ---------------------------------------------------------------------------

bool 
JavaPlugin::FileIsDirty(MProject& inProject, MFileRec &inFileRec, MakeStageT inStage, MakeActionT inAction, time_t inModDate)
{
	// See if inFile is newer than its output file
	
	BString outputFileName;
	makeOutputFileName(inFileRec.path, outputFileName);
	
	// assume the best...
	bool isDirty = false;
		
	BEntry outputFile(outputFileName.String());
	if (outputFile.Exists() == false || outputFile.IsFile() == false) {
		// if we can't find the output file or it isn't a file
		// we have to assume the source is dirty
DBG(("File %s doesn't exist\n", outputFileName.String() ));
		isDirty = true;
	}
	else {
		time_t outputModDate;
		outputFile.GetModificationTime(&outputModDate);
		if (outputModDate < inModDate) {
			// if the output file is older than our input file
			// it is dirty
DBG(("File %s is dirty\n", outputFileName.String() ));
			isDirty = true;
		}
	}
	return isDirty;	
}

// ---------------------------------------------------------------------------

void 
JavaPlugin::CodeDataSize(MProject& inProject, const char *inFilePath, int32 &outCodeSize, int32 &outDataSize)
{
	outCodeSize = -1;  // !!! FIXME
	outDataSize = -1;  // !!! FIXME
}

// ---------------------------------------------------------------------------

status_t 
JavaPlugin::GenerateDependencies(MProject& inProject, const char *inFilePath, BList &outList)
{
	// we are not dependent on anything, our output file is
	// but we aren't called for that one, the compiler is (or whatever tool
	// is used to include it in the build)
	return B_ERROR;
}

// ---------------------------------------------------------------------------

void 
JavaPlugin::GetTargetFilePaths(MProject& inProject, MFileRec &inFileRec, BList &inOutTargetFileList)
{
	// this is used by the linker to generate the list of object files to link
	// not applicable for tools

	//BString outputFileName;
	//makeOutputFileName(inFileRec.name, outputFileName);
	//inOutTargetFileList.AddItem(strdup(outputFileName.String()));
}


// ---------------------------------------------------------------------------

void 
JavaPlugin::makeOutputFileName(const char *filePath, BString &outputName)
{
	// take the input filePath and strip off the last .java and put on a .class
	outputName = filePath;
/**
	BPath file(filePath);
	outputName = projectPath.Path();
	outputName.Append("/");
	outputName.Append(file.Leaf());
**/
	outputName.ReplaceLast(".java", ".class");
DBG(("Output file path: %s\n", outputName.String() ));
}


// ---------------------------------------------------------------------------

const char kNewLine = '\n';

status_t 
JavaPlugin::ParseMessageText(MProject& inProject, const char *inText, BList &outList)
{	
	// put the text into a BString and make sure it ends with a newline
	// for less edge case testing below...
	BString text(inText);
	if (text[text.Length()-1] != kNewLine) {
		text += kNewLine;
	}

	int32 lineEnd = 0, tempEnd = 0;
	int32 lineStart = 0, tempStart = 0;
	int   lineCnt=0;
	
	switch( currentCompiler() )
	{
	case COMPILER_JAVAC: 
		DBG(("javac/pizza error reporting\n"));
		// iterate through the text 
		while ((lineEnd = text.FindFirst(kNewLine, lineStart)) != -1) {
			// skip empty lines
			if (lineEnd > lineStart + 1) {
			
				// skip 2 more lines
				tempStart = lineEnd+1;
				for(lineCnt=0; lineCnt<2; lineCnt++) {
					if ((tempEnd = text.FindFirst(kNewLine, tempStart)) == -1) break;
					tempStart = tempEnd + 1;
				}
								
//DBG(("[Err %ld -> %ld .. %ld]\n", lineStart, lineEnd, tempEnd));
				BString errLines;
				text.CopyInto(errLines, lineStart, lineEnd-lineStart);
				
				if (lineCnt<2) {	// !!! hack to filter warning and error counts
					bool isNWarn = (lineEnd-lineStart >= 8) 
								    && (errLines.FindFirst("warning", 0) != -1);
					bool isNErr  = (lineEnd-lineStart >= 6)
									&& (errLines.FindFirst("error", 0) != -1);
DBG(("%d lines, warn: %d, err: %d]\n", lineCnt, isNWarn, isNErr));
					if (! (isNWarn || isNErr))
						outList.AddItem( createTextOnlyMessage(errLines) );
				} else
					outList.AddItem(parseFileLineError(errLines, "", ":", "", ":"));
			}
			lineStart = max(lineEnd, tempEnd) + 1;
		}	
		break;
		
	case COMPILER_JIKES:
		DBG(("jikes error reporting\n"));
		ErrorMessage * error;
		int32 msgStart, msgEnd;
		// iterate through the text one line at a time 
		while ((lineEnd = text.FindFirst(kNewLine, lineStart)) != -1) {
			// skip empty lines
			if (lineEnd > lineStart + 1) {
				BString oneLine;
				
				// first line for pos. info
				text.CopyInto(oneLine, lineStart, lineEnd-lineStart);
				error = parseFileLineError(oneLine, "", ":", "", ":");
				// second line for message
				msgStart = lineEnd + 1;
				if ((msgEnd = text.FindFirst(kNewLine, msgStart)) != -1) {
					text.CopyInto(oneLine, msgStart, msgEnd-msgStart);
					DBG(("additional line:%s\n", oneLine.String()));
					if (oneLine[0]==' ' || oneLine[0]=='\t') {
						strncpy(error->errorMessage, oneLine.String(), MAX_ERROR_TEXT_LENGTH-1);
						lineEnd = msgEnd;
					}
				}	
				outList.AddItem(error);
			}
			lineStart = lineEnd + 1;
		}
		break;
		
	default:
		DBG(("generic error reporting\n"));
		// iterate through the text one line at a time 
		while ((lineEnd = text.FindFirst(kNewLine, lineStart)) != -1) {
			// skip empty lines
			if (lineEnd > lineStart + 1) {
				BString oneLine;
				
				text.CopyInto(oneLine, lineStart, lineEnd-lineStart);
// !!! FIXME    parseFileLineError chokes on guavac errors
//				outList.AddItem(parseFileLineError(oneLine, "", ":", "", ":"));
				outList.AddItem( createTextOnlyMessage(oneLine) );
			}
			lineStart = lineEnd + 1;
		}
	}
		
	return B_OK;
}

// ---------------------------------------------------------------------------

ErrorMessage*
JavaPlugin::createEmptyError()
{	
	ErrorMessage* error = new ErrorMessage;
	// -1 is a sentinal to trigger the creation of dynamic offsets
	// in MMessageItem.  (With an offset and length, the error message
	// will try to track edits made to the file.)
	error->offset = -1;
	error->length = 0;
	error->synclen = 0;
	error->syncoffset = 0;
	error->errorlength = 0;
	error->erroroffset = 0;
	
	error->filename[0] = 0;
	error->errorMessage[0] = 0;
	error->errorMessage[MAX_ERROR_TEXT_LENGTH-1] = 0;
	
	// the default (until we correctly parse the filename/linenumber) is text only
	error->linenumber = 0;
	error->textonly = true;
	error->isWarning = false;
	
	return error;
}

// ---------------------------------------------------------------------------

ErrorMessage*
JavaPlugin::createTextOnlyMessage(const BString& text)
{
	ErrorMessage* error = this->createEmptyError();
	int32 amount = min(text.Length(), MAX_ERROR_TEXT_LENGTH-1);
	text.CopyInto(error->errorMessage, 0, amount);
	error->errorMessage[amount] = 0;

	return error;
}

// ---------------------------------------------------------------------------

char *
JavaPlugin::getCompletePath(char * file)
{
int32 i;
MFileRec * rec;
int n = strlen(file);

	DBG(("looking for %s\n", file));
	for(i=fFiles->CountItems()-1;i>=0;i--) {
		rec = (MFileRec *) fFiles->ItemAt(i);
		if (rec) { 
			char * p = rec->path + strlen(rec->path) - n;
			DBG(("%s %s\n", p, rec->path)); 
			if (strcmp(p, file)==0) 
				return rec->path;
		} 			
	}
	return NULL;
}

// ---------------------------------------------------------------------------

// javac-style warning message
const char * WARN_STR = " warning:";

ErrorMessage*
JavaPlugin::parseFileLineError(const BString& text,
								const char* filePrefix, 
								const char* fileSuffix, 
								const char* linePrefix, 
								const char* lineSuffix)
{
	// Look for the following in the text:
	//	filePrefix<fileName>fileSuffix
	// 	linePrefix<lineNumber>lineSuffix
	// If we don't succeed in finding everything, create
	// a text only error message with the full text
	// as the error text
	
	// rather than deeply nesting, I return at text only
	// message at the first problem encountered...
	
	int32 fileStart = 0;
#if 0
	fileStart = text.FindFirst(filePrefix);
	if (fileStart == -1) {
		return this->createTextOnlyMessage(text);
	}
	fileStart += strlen(filePrefix);
#endif
DBG(("[file start %ld]", fileStart));

	int32 fileEnd = text.FindFirst(fileSuffix, fileStart);
	if (fileEnd == -1) {
		return this->createTextOnlyMessage(text);
	}
DBG(("[file end %ld]", fileEnd));
	
	int32 lineStart;
#if 0
	lineStart = text.FindFirst(linePrefix, fileEnd);
	if (lineStart == -1) {
		return this->createTextOnlyMessage(text);
	}
#else
	lineStart = fileEnd+1;
#endif
	lineStart += strlen(linePrefix);
DBG(("[line start %ld]", lineStart));

	int32 lineEnd = text.FindFirst(lineSuffix, lineStart);
	if (lineEnd == -1 /*|| text[lineStart]==' '*/) {
		return this->createTextOnlyMessage(text);
	}
DBG(("[line end %ld]", lineEnd));
	
	// we found the file and line
	// set up the file/line in our error message and
	// copy only the remaining text as the message
	ErrorMessage* error = this->createEmptyError();

	int32 length;

	// handle the file name
	length = min(fileEnd-fileStart, MAX_ERROR_PATH_LENGTH-1);
	text.CopyInto(error->filename, fileStart, length);
	error->filename[length] = 0;
	error->textonly = false;
	
	// try to recall entire pathname, if unique
	char * cPath = getCompletePath(error->filename);
	if (cPath)
		strncpy(error->filename, cPath, MAX_ERROR_PATH_LENGTH-1);
DBG((" File: %s", error->filename));

#if 1
	// ...the line number
	char lineNumberBuffer[32];
	length = min(lineEnd-lineStart, (int32) 32);
	text.CopyInto(lineNumberBuffer, lineStart, length);
	lineNumberBuffer[length] = 0;
	error->linenumber = atoi(lineNumberBuffer);
DBG((" Line: %ld\n", error->linenumber));

	// ...and the error message text
	int32 messageStart = lineEnd + strlen(lineSuffix);
	
	// ...is it a warning?
	int32 warningStart = text.IFindFirst(WARN_STR, messageStart);
	if (warningStart == messageStart)
	{ 
	  error->isWarning = true;
	  messageStart += strlen(WARN_STR);
	}
	// ...copy message text
	length = min(text.Length()-messageStart, MAX_ERROR_TEXT_LENGTH-1);
	text.CopyInto(error->errorMessage, messageStart, length);
	
#else
	// fallback : no parsing of message, just a copy 
	length=min(text.Length(), MAX_ERROR_TEXT_LENGTH-1);
	text.CopyInto(error->errorMessage, 0, length);
#endif
	error->errorMessage[length] = 0;
	return error;
}

