/**
 * @file evt.cpp 
 * Event system, handles translation from BeOS messages to Java events
 *
 * Copyright (c) 1998-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * @author T. Paoletti <tpaoletti@users.sourceforge.net>,
 *         <your name here>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 *
 *	@todo rewrite menu event dispatching code (a little obfuscated, right now)
 *  @todo redesign OnCommand() to suit BeOS event system
 */

#include "toolkit.hpp"

#include <MessageFilter.h>
#include <MessageQueue.h>

BMessage * KWM_WND_EVENT;
BMessage * KWM_REQUEST_FOCUS;
BMessage * KWM_CONTEXTMENU;
BMessage * KWM_DESTROY;

extern char * RootViewName;

#define EVENT_NULL		(uint32)0
#define EVENT_MENU		'KMnu'
#define EVENT_CONTROL	'KCmd'


#define BLOCKING		1

/*** debug ***/
#define VERBOSE_EVT		1

/*** local functions ***/
	/* the common event filters */
filter_result	windowEventFilter (BMessage * msg, BHandler ** target, BMessageFilter * msgFilter);
filter_result	viewEventFilter (BMessage * msg, BHandler ** target, BMessageFilter * msgFilter);

	/** Translate native BeOS events to Java events */
int TranslateEvent(BMessage * msg);

//------------------------------------------------------------------------------------
#ifndef D_LOOP_MODE

/** 
   A queue used to pass native events (messages) 
   from the native layer (windows' BLoopers and application event loop) to the AWT dispatcher thread

   With D_LOOP_MODE=1 we are posting event directly from the native layer to the Java event queue,
   so this is not needed. 
*/
class EventQueue {
	public:
		EventQueue( ) {
			queue = new BMessageQueue();
			eventReady= create_sem(1, "eventReady");	
			
			cntnull = 0;
			cnttot  = 0;
		}
		
		/** Post an event without blocking */
		int PostEvent(BMessage * msg, bool canskip) {
			
			if (queue->Lock()) {
					/* skip, if possible, annoying events during back-logs */
				if (!(canskip && (queue->CountMessages() > 10))) {	
					queue->AddMessage(msg);	
					DBG(AWT_EVT, ("queue: posted event %p [%d]\n", msg, queue->CountMessages() ));
				}
				queue->Unlock();
				release_sem(eventReady);
				return 1;
			} else
				return 0;
		}
		
		/** Poll for events
		  
		    It may or may not block depending on the BLOCKING define */
		BMessage *	GetEvent () {
			BMessage * msg;
			
#if BLOCKING
			acquire_sem(eventReady);
#endif
			if (queue->Lock()) {
				msg = queue->NextMessage();
				queue->Unlock();	
#if VERBOSE_EVT
				cnttot++;
				if (msg) {
					DBG(AWT_EVT, ("*** queue: got event %p [%d] (null %ld / tot. %ld ) \n", 
						msg, queue->CountMessages(),
						cntnull, cnttot ));
					DBG_ACTION(AWT_EVT, msg->PrintToStream());
				} else 
					cntnull++;
#endif
				return msg;
			} else 
				return NULL;
		}		
	
	private:
		BMessageQueue * queue;
		sem_id			eventReady;		
		int32			cntnull, cnttot;
};
#endif

/*******************************************************************************
*
*/

extern "C" {

	/** Initialize native event system */
	jobject __cdecl
		Java_java_awt_Toolkit_evtInit ( JNIEnv* env, jclass clazz )
	{
		jclass Component;
		
		DBG(AWT_EVT, ("evtInit\n"));
		
		ComponentEvent = env->FindClass( "java/awt/ComponentEvt");
		MouseEvent     = env->FindClass( "java/awt/MouseEvt");
		FocusEvent     = env->FindClass( "java/awt/FocusEvt");
		WindowEvent    = env->FindClass( "java/awt/WindowEvt");
		KeyEvent       = env->FindClass( "java/awt/KeyEvt");
		PaintEvent     = env->FindClass( "java/awt/PaintEvt");
		WMEvent        = env->FindClass( "java/awt/WMEvent");
		ItemEvent      = env->FindClass( "java/awt/ItemEvt");
		ActionEvent    = env->FindClass( "java/awt/ActionEvt");
		AdjustmentEvent= env->FindClass( "java/awt/AdjustmentEvt");
		
		getComponentEvent = env->GetStaticMethodID( ComponentEvent, "getEvent", 
			"(IIIIII)Ljava/awt/ComponentEvt;");
		getMouseEvent     = env->GetStaticMethodID( MouseEvent, "getEvent",
			"(IIIII)Ljava/awt/MouseEvt;");
		getFocusEvent     = env->GetStaticMethodID( FocusEvent, "getEvent",
			"(IIZ)Ljava/awt/FocusEvt;");
		getWindowEvent    = env->GetStaticMethodID( WindowEvent, "getEvent",
			"(II)Ljava/awt/WindowEvt;");
		getKeyEvent       = env->GetStaticMethodID( KeyEvent, "getEvent",
			"(IIIII)Ljava/awt/KeyEvt;");
		getPaintEvent     = env->GetStaticMethodID( PaintEvent, "getEvent",
			"(IIIIII)Ljava/awt/PaintEvt;");
		getWMEvent        = env->GetStaticMethodID( WMEvent, "getEvent",
			"(II)Ljava/awt/WMEvent;");
		getItemEvent      = env->GetStaticMethodID( ItemEvent, "getEvent",
			"(III)Ljava/awt/ItemEvt;");
		getActionEvent    = env->GetStaticMethodID( ActionEvent, "getEvent",
			"(ILjava/lang/String;I)Ljava/awt/ActionEvt;");
		getAdjustmentEvent= env->GetStaticMethodID( AdjustmentEvent, "getEvent",
			"(III)Ljava/awt/AdjustmentEvt;");
		
#if 0	/* !!! not needed */		
		dispatch = env->GetMethodID( WMEvent, "dispatch", "()V;" );

		/* keep for reference, window manager events with special functions */
		KWM_WND_EVENT = RegisterWindowMessage( TEXT("KWM_WND_EVENT") );
		KWM_REQUEST_FOCUS = RegisterWindowMessage( TEXT("KWM_REQUEST_FOCUS") );
		KWM_CONTEXTMENU= RegisterWindowMessage( TEXT("KWM_CONTEXTMENU") );
		KWM_DESTROY = RegisterWindowMessage( TEXT("KWM_DESTROY") );
#endif
	
		X->nWindows = 199;
		X->windows = (KWnd**)AWT_CALLOC( X->nWindows, sizeof(KWnd*));

		X->eventQueue = new EventQueue();
		
		Component = env->FindClass("java/awt/Component");
		return env->NewObjectArray( X->nWindows, Component, NULL);
	}
	
	
		/** Fetch events from the native layer.
		
			Called from within the AWT dispatcher thread.  */
	jobject __cdecl
		Java_java_awt_Toolkit_evtGetNextEvent ( JNIEnv* env, jclass clazz )
	{

#ifndef D_LOOP_MODE
		
//		DBG(AWT_EVT, ("evtGetNextEvent\n"));
/*		
		MSG msg;
		
		UNBLOCK_EXECUTE( clazz,
			if ( GetMessage( &msg, NULL, 0, 0) == true) {
				TranslateMessage( &msg);
				X->jEvt = NULL;
				DispatchMessage( &msg);
			}
		);
*/
		BMessage * msg;
		UNBLOCK_EXECUTE( clazz,

			X->jEvt = NULL;
						
			msg = X->eventQueue->GetEvent();
			if (msg!=NULL) {
//				msg->PrintToStream();
				TranslateEvent(msg);
				delete msg;
			} 					
		);
		return X->jEvt;
#else
		return NULL;
#endif
	}
	
		/** just a dummy for the linker */
	jobject __cdecl
		Java_java_awt_Toolkit_evtPeekEvent ( JNIEnv* env, jclass clazz )
	{
		return NULL;
	}
	
		/** just a dummy for the linker */
	jobject __cdecl
		Java_java_awt_Toolkit_evtPeekEventId ( JNIEnv* env, jclass clazz, jint id )
	{
		
		return NULL;
	}
	
		/** Force context switch to the event dispatcher thread */
	void __cdecl
		Java_java_awt_Toolkit_evtSendWMEvent ( JNIEnv* env, jclass clazz, jobject evt )
	{
		
#ifdef D_LOOP_MODE
		DBG(AWT_EVT, ("evtSendWMEvent\n"));
UNIMPLEMENTED((
		PostMessage( X->wakeUp, KWM_WND_EVENT, (WPARAM)evt, 0);
));
#endif
	}
	
		/** Wake up the event dispatcher thread */
	void __cdecl
		Java_java_awt_Toolkit_evtWakeup ( JNIEnv* env, jclass clazz )
	{
#ifndef D_LOOP_MODE
/*		if ( X->dspTid )
			PostThreadMessage( X->dspTid, WM_NULL, 0, 0);
*/
		if (X && X->eventQueue) {
			X->eventQueue->PostEvent(new BMessage(EVENT_NULL), false);
		}
#endif
	}
	
		/** Insert a KWnd widget among the event handlers */
	jint __cdecl
		Java_java_awt_Toolkit_evtRegisterSource ( JNIEnv* env, jclass clazz, KWnd* wnd )
	{
		register unsigned int i;
		int      n;
		
		/*
		* we don't use a double hashing here because collisions are very unlikely
		* (window IDs usually already are hashed, so it does not make sense to
		* hash them again - we just could make it worse
		*/
		for ( i = (unsigned long)wnd, n=0; n < X->nWindows; i++, n++ ) {
			i %= X->nWindows;
			if ( (int)X->windows[i] <= 0 ) {
				X->windows[i] = wnd;
				X->windows[i]->srcIdx = i;
/* !!!
				setWnd( X->windows[i]->getWindow(), wnd);
*/
				DBG(AWT_EVT, ("evtRegisterSource: %s %p -> %d\n", wnd->getView()->Name(), wnd, i));
				
				/* attach event filter */
				if (wnd->lock()) {
					wnd->getView()->AddFilter(new BMessageFilter(B_ANY_DELIVERY, B_ANY_SOURCE, viewEventFilter));
					wnd->unlock();
				}
				
				return i;
			}
		}
		
		fprintf(stderr, "window table out of space: %d", X->nWindows);
		return -1;
	}
	
		/** Remove widget from the event handlers */
	jint __cdecl
		Java_java_awt_Toolkit_evtUnregisterSource ( JNIEnv* env, jclass clazz, KWnd* wnd )
	{
		int i = wnd->srcIdx;

		DBG(AWT_EVT, ("evtUnregisterSource\n"));
		/*
		* native destruction has to be async by creator thread.
		* window objects and childs are deleted by wndDestroyWindow
		* WM_DESTROY event processing
		*/
		
		if ( i >= 0 ) {
			X->windows[i] = (KWnd*)-1;
			return i;
		}
		else {
			return -1;
		}
	}

};	/* end of extern "C" */

class BView;

/*--------------------------------------------------------------
   code to inspect the entire hierarchy of views
*/
void recurseView(BView * v, BMenu * menu) {
	int cnt = v->CountChildren();
	
	char info[200];
	BRect r = v->Frame();
	rgb_color bc = v->ViewColor();
	rgb_color fc = v->HighColor();
	
	sprintf(info, "%s %p at %.0f,%.0f,%.0f,%.0f %x %x %s", 
		v->Name(), 
		v, r.left, r.top, r.right, r.bottom,
		*(long *)&fc, *(long *)&bc, 
		(v->IsHidden() ? "hidden" : "")); 
		
	if (cnt == 0) {
		menu->AddItem(new BMenuItem(info, NULL));
	} else {
		BView * subv;	
		BMenu * submenu = new BMenu(info);
	
		for(int i=0; i<cnt; i++) {
			subv = v->ChildAt(i);		
			recurseView(subv, submenu);
		}
		menu->AddItem(submenu);
	}
}

/** Shows a popup menu which represents the hierarchy of native widgets. 

    The popup is activated clicking on a view with the CTRL modifier pressed.
    Called from viewEventFilter() and windowEventFilter() for debugging purposes.
*/
int32 InspectHierarchy(BMessage * msg )
{				
#ifdef DEBUG
	char info[200];
	BView * v;
	BPoint where   = msg->FindPoint("where");
	BWindow * root;
	
	if (msg->FindPointer("root", (void **)&root) != B_OK) return -1; 
		
	BPopUpMenu * popup   = new BPopUpMenu("inspector");
	BMenuItem * debugEvt = new BMenuItem("Debug evt", NULL);
	BMenuItem * debugFnt = new BMenuItem("Debug fnt", NULL);
	BMenuItem * debugGra = new BMenuItem("Debug gra", NULL);
	BMenuItem * debugImg = new BMenuItem("Debug img", NULL);
	BMenuItem * debugWnd = new BMenuItem("Debug wnd", NULL);
			
	/* populate the menu */
	if (root->Lock()) {
			/* the window */
		BRect r = root->Frame();
		sprintf(info, "window %p at %.0f,%.0f,%.0f,%.0f", root, r.left, r.top, r.right, r.bottom); 
		popup->AddItem(new BMenuItem(info, NULL));
	
			/* all views by hierarchical order */
		for(int i=0; i<root->CountChildren(); i++) {
			v = root->ChildAt(i);
			recurseView(v, popup);
		}
			/* optional items (debug options, ...) */
		popup->AddItem(new BSeparatorItem);
		popup->AddItem(debugEvt);
		popup->AddItem(debugFnt);
		popup->AddItem(debugGra);
		popup->AddItem(debugImg);
		popup->AddItem(debugWnd);
		
		root->Unlock();			
	}
		/* show popup */
	BMenuItem * mi = popup->Go(where);
		/* manage options */
	if (mi==debugEvt) kaffevmDebugMask ^= DBG_AWT_EVT;
	if (mi==debugFnt) kaffevmDebugMask ^= DBG_AWT_FNT;
	if (mi==debugGra) kaffevmDebugMask ^= DBG_AWT_GRA;
	if (mi==debugImg) kaffevmDebugMask ^= DBG_AWT_IMG;
	if (mi==debugWnd) kaffevmDebugMask ^= DBG_AWT_WND;
		
	delete popup;
	
#endif
	return 0;
}

//------------------------------------------------------------------------------------

/** Get the KWnd object for target
*/	
KWnd * getWnd( BHandler * target )
{
	KWnd * wnd;

	/** @todo FIX, and once fixed move to toolkit.hpp as an inline 
		this slowly scans the list of registered event sources in search of this target 
	*/
	for(int i=0;i<X->nWindows;i++) {
		wnd = X->windows[i];
		if ((wnd>0) && (wnd->getView() == target)) {
			return wnd;
		}
	}
	return NULL;
}

//------------------------------------------------------------------------------------
/** The filter function for window-related events

  We don't want to break the native event chain, we just insert ourselves between
  message reception and message dispatch to the native objects

  message loop -> [our MessageFilter] -> DispatchMessage -> object-oriented handlers
*/
filter_result	windowEventFilter (BMessage * msg, BHandler ** target, BMessageFilter * msgFilter) 
{		
	KWnd * wnd;
	bool   canskip = false;
	
	BWindow * ourWindow = (BWindow *)msgFilter->Looper();

	if ((msg==NULL) || (target==NULL) || (*target!=ourWindow)) 
		return B_DISPATCH_MESSAGE;
		
	BView * ourRoot = ourWindow->FindView(RootViewName);

	/* identify the target - the root view for this window*/
	if ( ! (wnd = getWnd( ourRoot))) {
		return B_DISPATCH_MESSAGE;	
	}

	switch(msg->what) {
			// forwarded events
		case B_WINDOW_ACTIVATED:
		case B_MINIMIZE:
		case B_QUIT_REQUESTED:
			break;
			
		case EVENT_CONTROL:
		case EVENT_MENU:
			break;			/* !!! notifications */
		
		case _UPDATE_:
			canskip = true;
			break;

		case B_WINDOW_MOVED:
		case B_WINDOW_RESIZED:
			canskip = true;
			break;
	
		default:
			return B_DISPATCH_MESSAGE;
	}
	
		/* print almost all posted messages */
	if ((msg->what != B_MOUSE_MOVED) && (msg->what != _EVENTS_PENDING_)) {
		DBG_ACTION(AWT_EVT,  msg->PrintToStream() );
		DBG(AWT_EVT, ("  msg %p to %s %p (native %p)\n", 
						msg, (wnd ? wnd->getView()->Name() : ""), 
						wnd, *target));
	}

		/* post message directly to the event dispatcher */
	BMessage * outmsg = new BMessage(*msg);
	outmsg->AddPointer("KWnd", (void *)wnd);
	outmsg->AddBool("Kroot", true);
	X->eventQueue->PostEvent(outmsg, canskip);

	return B_DISPATCH_MESSAGE;
}


//------------------------------------------------------------------------------------
/** The filter function for view-related events

  We don't want to break the native event chain, we just insert ourselves between
  message reception and message dispatch to the native objects

  message loop -> [our MessageFilter] -> DispatchMessage -> object-oriented handlers
  
  This filter is added on view registration (see evtRegisterSource)
*/
filter_result	viewEventFilter (BMessage * msg, BHandler ** target, BMessageFilter * msgFilter) 
{		
	bool	canskip = false;
	KWnd * wnd;
			
	if (msg==NULL || target==NULL) return B_DISPATCH_MESSAGE;
	
	/* identify the target */
	if ( ! (wnd = getWnd(*target))) {
		return B_DISPATCH_MESSAGE;	
	}

	/*--- view inspector ---*/
#ifdef DEBUG
	if ((msg->what == B_MOUSE_DOWN) && (msg->FindInt32("modifiers") & B_LEFT_CONTROL_KEY)) {
		msg->AddPointer("root", wnd->getWindow());
		
			/* we spawn a thread because we don't want to interrupt the event chain */
		thread_id inspector = spawn_thread(InspectHierarchy, "inspector", B_DISPLAY_PRIORITY, msg);
		resume_thread(inspector);
		
		return B_SKIP_MESSAGE;	/* do not handle the mouse click used to activate this */		
	}
#endif
 	/*----------------------*/
		
		/* post message directly to the event dispatcher */
	switch(msg->what) {
			// forwarded events
		case B_MOUSE_DOWN: 
		case B_MOUSE_UP: 
		case B_KEY_DOWN: 
		case B_KEY_UP:
			break;
		case EVENT_CONTROL:	/* !!! window or view events ? */
		case EVENT_MENU:
			break;
		case B_MOUSE_MOVED:
			canskip = false; // !!! set to true
			break;
		default:
			return B_DISPATCH_MESSAGE;	/* everything else is dispatched natively */
	}

	/* print almost all posted messages */
	if ((msg->what != B_MOUSE_MOVED) && (msg->what != _EVENTS_PENDING_)) {
		DBG_ACTION(AWT_EVT, msg->PrintToStream() );
		DBG(AWT_EVT, ("  msg %p to %s %p (native %p)\n", 
						msg, (wnd ? wnd->getView()->Name() : ""), 
						wnd, *target));
	}
	
		/* post message directly to the event dispatcher */
	BMessage * outmsg = new BMessage(*msg);
	outmsg->AddPointer("KWnd", (void *)wnd);
	X->eventQueue->PostEvent(outmsg, true);
	
	return B_DISPATCH_MESSAGE;
}


//------------------------------------------------------------------------------------
/** Native-to-Java event translation

    This is the core routine of the event system,
    if some event doesn't get to the proper target, 
    please look here or into viewEventFilter() and windowEventFilter()
*/
int TranslateEvent(BMessage * msg) {
	KWnd * wnd = NULL;

	if (msg->what == EVENT_NULL) return 0;	/* just used to wake the event dispatcher */
						
	if (msg->FindPointer("KWnd", (void **)&wnd) != B_OK) {
		DBG(AWT_EVT, ("evt: no KWnd field\n"));
		return 0;	
	}	

	if (wnd)
	switch(msg->what) {
			// mouse
		case B_MOUSE_DOWN: 
			{
				int32 buttons= msg->FindInt32("buttons");
				BPoint where = msg->FindPoint("where");
				if ( wnd->OnMouseDown( buttons, where.x, where.y ) ) 
					return 0;
			}
			break;
		case B_MOUSE_UP:
			{
				int32 buttons= msg->FindInt32("buttons");
				BPoint where = msg->FindPoint("where");
				
				/** @todo TEST ONLY, REMOVE! */
				BRect upd (0,0,600,600);	
				if ( (modifiers() & B_SHIFT_KEY) && wnd->OnPaint( upd.left, upd.top, upd.Width(), upd.Height()) ) 
					return 0;
					
				if ( wnd->OnMouseUp( buttons, where.x, where.y ) ) 
					return 0;
					
					
			}
			break;
			
			// mouse motion
		case B_MOUSE_MOVED:
			{
				BPoint where  = msg->FindPoint("where");
				int32 transit = msg->FindInt32( "be:transit");

				if ( transit == B_ENTERED_VIEW) {
					if (wnd->OnMouseEnter( (int)where.x, (int)where.y) )
						return 0;
				}
				else
				if ( transit == B_EXITED_VIEW) {
					if (wnd->OnMouseExit( (int)where.x, (int)where.y) )
						return 0;
				}
				else
					if ( wnd->OnMouseMove( (int)where.x, (int)where.y) )
						return 0;
			}
			break;	

			// keyboard
		case B_KEY_DOWN:
			{
				const char * bytes;
				if (msg->FindString("bytes", &bytes) == B_OK)
					if ( wnd->OnKeyDown( bytes[0], 0) )
						return 0;
			}
			break;
		case B_KEY_UP:
			{
				const char * bytes;
				if (msg->FindString("bytes", &bytes) == B_OK)
					if ( wnd->OnKeyUp( bytes[0], 0) )
						return 0;
			}
			break;

			// notifications
		case EVENT_CONTROL:
			msg->PrintToStream();
			if ( wnd->OnCommand( 0))	/* KButton::OnCommand is called here */
				return 0;
			/*
			else {
				cwnd = (KWnd*)GetWindowLong( (HWND)lParam, GWL_USERDATA);
				if ( cwnd && cwnd->OnCommand( HIWORD( wParam) ))
					return 0;
			}
			*/
			break;
		case EVENT_MENU:	
			{			
				int32 id = msg->FindInt32("Kid");
				BMenu * source;
				if (msg->FindPointer("source", (void **)&source) == B_OK &&
					wnd->processMenuAction( id))		/* KWindow::OnCommand is not called ! */
					return 0;
			}
			break;

			// painting
		case _UPDATE_:
			{
				/** @todo GET REAL UPDATE AREA */
				BRect upd (0,0,600,600);	
				if ( wnd->OnPaint( upd.left, upd.top, upd.Width(), upd.Height()) ) 
					return 0;
			}			
			break;

			// expose
#if 0
		case _UPDATE_: /*WM_DRAWITEM:*/
			//cwnd = (KWnd*)GetWindowLong( GetDlgItem( hwnd, (int)wParam), GWL_USERDATA);
			if ( wnd->OnOwnerDraw( ) )
				return 0;
			break;
#endif
	
			// sizeing
#if 1	/* !!! had problems with size events */
		case B_WINDOW_MOVED:
			{	BPoint where = msg->FindPoint("where");
				BRect  fr = wnd->getWindow()->Frame();
				if ( wnd->OnPosChanged( where.x, where.y, fr.Width(), fr.Height()) )
					return 0;
			}
			break;
		case B_WINDOW_RESIZED:
			{	int32 w = msg->FindInt32("width");
				int32 h = msg->FindInt32("height");	
				BRect  fr = wnd->getWindow()->Frame();
				if ( wnd->OnPosChanged( fr.left, fr.top, w, h) )
					return 0;
			}
			break;
#endif

			// window management
		case B_WINDOW_ACTIVATED:
			{	bool active = msg->FindBool("active");
				if ( wnd->OnFocusChange( active) )
					return 0;
			}
			/* this does not map to a Java event
				if ( wnd->OnActivate( 0, 0, wnd->getWindow()) ) 
				return 0;
			*/
			break;
		
		case B_QUIT_REQUESTED:
			if (wnd->OnClose())
				return 0;
			break;
	}

	return 0;
}

//------------------------------------------------------------------------------------
#if WIN32_EVENT_HANDLER

	LRESULT WINAPI WndProc( HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) {
		LPWINDOWPOS pos;
		RECT rcl;
		KWnd *wnd, *cwnd;
		HBRUSH brush;
		LRESULT ret;

		DBG(AWT_EVT, ("hwnd:%x message:%x wparam:%x laram:%x \n", hwnd, iMsg, wParam, lParam));
//		printf("hwnd:%x message:%x wparam:%x laram:%x \n", hwnd, iMsg, wParam, lParam);
		
		/*
		* KWM events redirected to create thread
		*/
		//<-X->wakeUp
		if ( iMsg == KWM_WND_EVENT ) {
			JniEnv->CallVoidMethod( (jobject)wParam, dispatch);
			return 0;
		}
		//<-cmnSetFocus
		if ( iMsg == KWM_REQUEST_FOCUS ) {
			SetFocus( (HWND)wParam);
			return 0;
		}
		//<-cmn/wndDestroyWindow
		if ( iMsg == KWM_DESTROY ) {
			DestroyWindow( (HWND)wParam);
			return 0;
		}
		//<-menuAssign
		if ( iMsg == KWM_CONTEXTMENU ) {
			if ( (wnd = getWnd( hwnd)) )
				wnd->OnContextMenu( LOWORD(lParam), HIWORD(lParam) );
			return 0;
		}
		
		if ( ! (wnd = getWnd( hwnd)) ) {
			if ( IsWindow( hwnd) ) {
				return DefWindowProc( hwnd, iMsg, wParam, lParam);
			}
			else {
				return 0;
			}
		}
		
		switch( iMsg) {
			
		// expose
		case WM_DRAWITEM:
			cwnd = (KWnd*)GetWindowLong( GetDlgItem( hwnd, (int)wParam), GWL_USERDATA);
			if ( cwnd->OnOwnerDraw( (LPDRAWITEMSTRUCT) lParam) ) {
				return 0;
			}
			break;
			
			// colors
		case WM_CTLCOLOREDIT:
		case WM_CTLCOLORBTN:
		case WM_CTLCOLORLISTBOX:
		case WM_CTLCOLORSCROLLBAR:
		case WM_CTLCOLORSTATIC:
			cwnd = (KWnd*)GetWindowLong( (HWND)lParam, GWL_USERDATA);
			if ( (cwnd) && (brush = cwnd->OnCtlColor( (HDC)wParam)) ) {
				return (LRESULT)brush;
			}
			break;
			
			// painting
		case WM_ERASEBKGND:
			if ( wnd->OnEraseBkgnd( (HDC)wParam) ) {
				return (LRESULT)1;
			}
			break;
			
		case WM_PAINT:
			if ( GetUpdateRect( hwnd, &rcl, false)) {
				if ( wnd->OnPaint( rcl.left, rcl.top, rcl.right-rcl.left, rcl.bottom-rcl.top) ) {
					return 0;
				}
			}
			break;
			
			// keyboard
		case WM_SYSKEYDOWN:
		case WM_KEYDOWN:
			if ( wnd->OnKeyDown( (int)wParam, (int)lParam) )
				return 0;
			break;
		case WM_SYSKEYUP:
		case WM_KEYUP:
			if ( wnd->OnKeyUp( (int)wParam, (int)lParam) )
				return 0;
			break;
		case WM_CHAR:
			if ( wnd->OnChar( (TCHAR)wParam, (int)lParam) )
				return 0;
			break;
			
			// notifications
		case WM_COMMAND:
			//menu notification
			if ( lParam == 0) {
				if ( wnd->OnCommand( LOWORD( wParam) ))
					return 0;
			}
			else {
				cwnd = (KWnd*)GetWindowLong( (HWND)lParam, GWL_USERDATA);
				if ( cwnd && cwnd->OnCommand( HIWORD( wParam) ))
					return 0;
			}
			break;
		case WM_HSCROLL:
			cwnd = (KWnd*)GetWindowLong( (HWND)lParam, GWL_USERDATA);
			if ( cwnd && cwnd->OnScroll( HIWORD( wParam), LOWORD( wParam), false ))
				return 0;
			break;
		case WM_VSCROLL:
			cwnd = (KWnd*)GetWindowLong( (HWND)lParam, GWL_USERDATA);
			if ( cwnd && cwnd->OnScroll( HIWORD( wParam), LOWORD( wParam), true ))
				return 0;
			break;
			
			// sizeing
		case WM_WINDOWPOSCHANGED:
			pos = (LPWINDOWPOS) lParam;
			if ( wnd->OnPosChanged( pos->x, pos->y, pos->cx, pos->cy) ) {
				return 0;
			}
			break;
			
			// mouse motion
		case WM_MOUSEMOVE:
			if ( wnd->OnMouseMove( LOWORD( lParam), HIWORD( lParam)) ) {
				return 0;
			}
			break;
			
			// mouse
		case WM_LBUTTONDOWN:
		case WM_MBUTTONDOWN:
		case WM_RBUTTONDOWN:
			if ( wnd->OnMouseDown( (int)wParam,LOWORD( lParam), HIWORD( lParam) ) ) {
				return 0;
			}
			break;
		case WM_LBUTTONUP:
		case WM_MBUTTONUP:
		case WM_RBUTTONUP:
			if ( wnd->OnMouseUp( (int)wParam,LOWORD( lParam), HIWORD( lParam) ) ) {
				return 0;
			}
			break;
			
#if !defined(UNDER_CE)
		case WM_SETCURSOR:
		/*
		* do not set when entering menu mode or hittest
		* does not equal client
			*/
			if ( (HIWORD( lParam) != 0) && (LOWORD( lParam) == 1) ) {
				cwnd = (KWnd*)GetWindowLong( (HWND)wParam, GWL_USERDATA);
				if ( cwnd && cwnd->OnSetCursor() )
					return (LRESULT) 1;
			}
			break;
#endif
			// focus
		case WM_SETFOCUS:
			if ( wnd->OnFocusChange( true) ) {
				return 0;
			}
			break;
		case WM_KILLFOCUS:
			if ( wnd->OnFocusChange( false) ) {
				return 0;
			}
			break;
		case WM_ACTIVATE:
			if ( wnd->OnActivate( LOWORD( wParam), (bool)HIWORD( wParam), (HWND) lParam) ) {
				return 0;
			}
			break;

		case WM_CLOSE:
			if ( wnd->OnClose() ) {
				return 0;
			}
			break;
			
		case WM_CREATE:
			break;
			
		case WM_DESTROY:
			// unlink java objects
			SetWindowLong( wnd->hwnd, GWL_USERDATA, 0);
			ret = CallWindowProc( wnd->defWndProc, hwnd, iMsg, wParam, lParam);
			
			// cleanup
			X->lastHwnd = (HWND)-1;
			delete( wnd);
			
			return ret;
			
		default:
			break;
		}
		
		return CallWindowProc( wnd->defWndProc, hwnd, iMsg, wParam, lParam);
	}
	
#endif /* WIN32_EVENT_HANDLER */

