/**
 * @file fnt.cpp
 * Font system. 
 *
 * Copyright (c) 1999-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * Portions Copyright (c) 1999
 *      Sim IJskes
 *
 * @author Sim IJskes <sim@nyx.xs4all.nl>
 *         T. Paoletti <tpaoletti@users.sourceforge.net>
 *         <your name here>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include <string.h>
#include "toolkit.hpp"

/* snippet from Font.java (kaffe-1.0b4)
       final public static int PLAIN = 0;
        final public static int BOLD = 1;
        final public static int ITALIC = 2;
*/

#define BOLD 1
#define ITALIC 2 

/** 
	The font system core routine,
    translates font specs into native BFont objects
		
	@todo FULLY IMPLEMENT STYLES
*/
BFont *findFont( const char *spec, int style )
{
	if( strcmp( spec, "Dialog" ) == 0 ) {
		return new BFont( be_plain_font );
	}

	if( strcmp( spec, "Default" ) == 0 ) {
		// is this conform original java spec?
		return new BFont( be_plain_font );
	}

	if( strcmp( spec, "Monospaced" ) == 0 ) {
		return new BFont( be_fixed_font );
	}

	if( strcmp( spec, "Serif" ) == 0 ) {
		TOFIX();
		return new BFont( be_fixed_font );
	}

	if( strcmp( spec, "SansSerif" ) == 0 ) {
		TOFIX();
		return new BFont( be_fixed_font );
	}

	TOFIX();

#ifdef DEBUG
	int32 numFamilies = count_font_families();

	for ( int32 i = 0; i < numFamilies; i++ ) {
		font_family family;
		uint32 flags;
		if ( get_font_family(i, &family, &flags) == B_OK ) {

			printf( "have font %s\n", family );

			int32 numStyles = count_font_styles(family);
			for ( int32 j = 0; j < numStyles; j++ ) {
				font_style style;
				if ( get_font_style(family, j, &style, &flags) == B_OK ) {
					printf( "   style '%s'\n", style );
				}
			}
		}
	}
#endif
	return new BFont(be_fixed_font);
}


extern "C" {
	
	void*
		Java_java_awt_Toolkit_fntInitFont ( JNIEnv* env, jclass clazz, jstring jSpec, jint style, jint size )
	{
		char *spec = java2CString( env, X, jSpec);

		DBG(AWT_FNT, ( "fntInitFont asked '%s' style=%x \n", spec, style ));	

		BFont* fs = findFont( spec, style );
		fs->SetSize(size);

		DBG_ACTION(AWT_FNT, fs->PrintToStream());

		return (void*) fs; 
	}
	
	void
		Java_java_awt_Toolkit_fntFreeFont ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		delete pFont;
	}
	
	/*******************************************************************************
	* FontMetrics support
	*/
	
	void*
		Java_java_awt_Toolkit_fntInitFontMetrics ( JNIEnv* env, jclass clazz, Font* pFont )
	{
/* !!!
		SelectObject( X->display, pFont->fnt);
		GetTextMetrics( X->display, &(pFont->fm));
*/		
		return pFont;
	}
	
	void
		Java_java_awt_Toolkit_fntFreeFontMetrics ( JNIEnv* env, jclass clazz, Font* pFont )
	{
	}
	
	
	jint
		Java_java_awt_Toolkit_fntGetAscent ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		font_height hgt;
		pFont->GetHeight(&hgt);
		return (jint) hgt.ascent;
	}
	
	jint
		Java_java_awt_Toolkit_fntGetDescent ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		font_height hgt;
		pFont->GetHeight(&hgt);
		return (jint) hgt.descent;	}
	
	jint
		Java_java_awt_Toolkit_fntGetFixedWidth ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		//		return pFont->fm.tmMaxCharWidth;

		if (pFont->IsFixed()==0)
			return 0;

		int w = (int)pFont->StringWidth( "H" );
		DBG(AWT_FNT, ( "getFixedWidth=%d\n", w ));
		return (jint)w ; 
	}
	
	jint
		Java_java_awt_Toolkit_fntGetHeight ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		font_height hgt;
		pFont->GetHeight(&hgt);
		return (jint)( hgt.ascent + hgt.descent + hgt.leading );
	}
	
	jint
		Java_java_awt_Toolkit_fntGetLeading ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		font_height hgt;
		pFont->GetHeight(&hgt);
		return (jint) hgt.leading;	}
	
	jint
		Java_java_awt_Toolkit_fntGetMaxAdvance ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		// conform spec. 'if not known the function should return -1'
		return -1 ;
	}
	
	jint
		Java_java_awt_Toolkit_fntGetMaxAscent ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		/** @todo CHECK, IS IT THE MAX ASCENT ? */
		font_height hgt;
		pFont->GetHeight(&hgt);
		return (jint) hgt.ascent;	
	}
	
	jint
		Java_java_awt_Toolkit_fntGetMaxDescent ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		/** @todo CHECK, IS IT THE MAX DESCENT ? */
		font_height hgt;
		pFont->GetHeight(&hgt);
		return (jint) hgt.descent;
	}
	
	jboolean
		Java_java_awt_Toolkit_fntIsWideFont ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		//		return ((pFont->fm.tmPitchAndFamily & 0xf0) == TMPF_VECTOR);
		
		return FALSE;
	}
	
	jobject
		Java_java_awt_Toolkit_fntGetWidths ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		int       n = 256;
		jarray    widths;
		jint      *jw;
		jboolean isCopy;
		register  int i; 
		BFont *f = pFont ;
	  
		widths = env->NewIntArray( n );
		
		DBG(AWT_FNT, ("fntGetWidths %x\n", pFont));
		
		jw = env->GetIntArrayElements( widths, &isCopy);
	
		for(i=0;i<n;i++) {
			char buf[] = "X" ;
			*buf = i ;
			jw[i] = (int)f->StringWidth( buf );
		}
	
		env->ReleaseIntArrayElements( widths, jw, 0);
	
		return widths;	
	}
	
	
	jint
		Java_java_awt_Toolkit_fntBytesWidth ( JNIEnv* env, jclass clazz, Font* pFont,
		jarray jBytes, jint off, jint len )
	{
/* !!!
		SIZE sz;
		jboolean  isCopy;
		jbyte    *jb = env->GetByteArrayElements( jBytes, &isCopy);
		int       n = env->GetArrayLength( jBytes);
		TCHAR	 *cs;
		
		if ( off+len > n ) len = n - off;
		
		cs = jchar2CString( X, (jchar*)jb+off, len);
		SelectObject( X->display, pFont->fnt);
		GetTextExtentExPoint( X->display, cs, len, 0, NULL, NULL, &sz);
		
		env->ReleaseByteArrayElements( jBytes, jb, JNI_ABORT);
		return sz.cx;
*/
		jboolean  isCopy;
		jbyte    *jb = env->GetByteArrayElements( jBytes, &isCopy);
		int       n = env->GetArrayLength( jBytes);
		int       w;

		DBG(AWT_FNT, ( "should not be called, because isWideFont is false\n" ));

		if ( off+len > n ) len = n - off;

		w = (int) (pFont->StringWidth(jb+off, len));

		env->ReleaseByteArrayElements( jBytes, jb, JNI_ABORT);

		return w; 
	}
	
	jint
		Java_java_awt_Toolkit_fntCharWidth ( JNIEnv* env, jclass clazz, Font* pFont, jchar jChar )
	{
UNIMPLEMENTED((
#ifndef WORDS_BIGENDIAN
		jChar = (jChar << 8) | (jChar >> 8);
#endif
));
		DBG(AWT_FNT, ("font:%p", pFont));
		return (jint) pFont->StringWidth((char *)&jChar, 1);
	}
	
	jint
		Java_java_awt_Toolkit_fntCharsWidth ( JNIEnv* env, jclass clazz, Font* pFont,
		jarray jChars, jint off, jint len )
	{
		jboolean  isCopy;
		jchar    *jc = env->GetCharArrayElements( jChars, &isCopy);
		int      n = env->GetArrayLength( jChars);
		char     *b;
		int      w;

		if ( off+len > n ) len = n - off;

UNIMPLEMENTED((

#ifndef WORDS_BIGENDIAN
		n = sizeof(XChar2b)*len;
		b = (XChar2b*) getBuffer( Tlk, n);
		swab( (jc+off), b, n);
#else
		b = (XChar2b*) (jc + off);
#endif
));
		b = (char *)(jc + off);
		w = (jint) pFont->StringWidth(b, len);

		env->ReleaseCharArrayElements( jChars, jc, JNI_ABORT);
		return w;
	}
	
	jint
		Java_java_awt_Toolkit_fntStringWidth ( JNIEnv* env, jclass clazz, Font* pFont, jstring jStr )
	{
		jboolean isCopy;
		const jchar    *jc = env->GetStringChars( jStr, &isCopy);
		int      len = env->GetStringLength( jStr);
		int      w, n;
		char     *b;
  
		/** @todo BAD IMPLEMENTATION, TO BE IMPROVED */
		b = (char *)jc;  
		w=0;
		for(n=0;n<len;n++, b+=2)
 			w += (jint) pFont->StringWidth(b, 1);
 
 		env->ReleaseStringChars( jStr, jc);
		return w;

/** @todo REVIEW AND CORRECT
		SIZE sz;
		TCHAR *jc = java2WinString( env, X, jStr);
		
		SelectObject( X->display, pFont->fnt);
		GetTextExtentExPoint( X->display, jc, _tcslen(jc), 0, NULL, NULL, &sz);
		
		return sz.cx;
*/
	}
	
}
