/**
 * @file img.c
 * Image support. 
 *	
 * Uses Translation Kit to load different image formats like JPEG, GIF and PNG.
 *
 * Copyright (c) 1998-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * @author T. Paoletti <tpaoletti@users.sourceforge.net>,
 *         <your name here>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 	
 *  @todo IMPLEMENT imgProduceImage()
 *  @todo image rings (GIF movies)
 *  @todo alpha channel reduction
 *  @todo support for GIF transparency (see GIFTranslator documentation)
 */

#include <StorageKit.h>
#include <TranslationKit.h>

#include "toolkit.hpp"

extern char * RootViewName;

/************************************************************************************
 * own auxiliary funcs
 */


/** Create a new image, with an optional image as background */ 
Image*
createImage ( int width, int height, Image* background )
{
BBitmap * bmap;

	DBG( AWT_IMG, ( "createImage: %d %d (bgnd: %p)\n", width, height, background));

	Image * img = (Image *) AWT_CALLOC( 1, sizeof( Image));
	if (img==NULL) return NULL;

			/* !!! A TEST */
	if (width==0 && height==0) {
		width = 300;
		height= 300;
	}

	img->trans  = -1;     /* default to no alpha */
	img->width = width;   /* we need to (temp) store them for subsequent image creation */
	img->height = height;
	
	/* create bitmap */
	bmap = new BBitmap(BRect(0,0,width,height), B_RGB_32_BIT, true);   
	BView   * tmp  = new BView(BRect(0,0,width,height), RootViewName, B_FOLLOW_ALL_SIDES, 0);
	
	/* paint into and sync */
	if (bmap != NULL && bmap->Lock()) {
		bmap->AddChild(tmp);
		tmp->StrokeRect(BRect(2,2,width-4,height-4));	/* DEBUG */
		if (background!=NULL) {
    			tmp->DrawBitmap(background->bitmap, BRect(0,0,width,height));
		}
		tmp->Sync();
		bmap->Unlock();
	} else { 
/* !!!
		DBG(AWT_IMG, ("createImage failed\n"));
		freeImage(X, img);
		return NULL;
*/	}

	img->bitmap = bmap; 

	return img;
}

int32 pixelValue(Toolkit * X, int32 rgb)
{
	return rgb;		// !!! see if we need color conversion here
}


void freeImage(Toolkit * X, Image* img) 
{
	if (img!=NULL) {
		/* delete native objects */
		if (img->bitmap) {
			delete img->bitmap;
		}
		AWT_FREE( img);
	}

UNIMPLEMENTED((
  Image *next, *first = img;
  
  /* we have to be aware of image rings (GIF movies), iterate */
  do {
	if ( img->pix ){
	  XFreePixmap( X->dsp, img->pix);
	  img->pix = 0;
	}

	/*
	 * note that XDestroyImage automatically frees any non-NULL data pointer
	 * (since we explicitly allocated the data, we better free it explicitly, too)
	 * malloc, free might be resolved
	 */

	if ( img->xImg ){
	  if ( img->shmiImg ) {
		DBG( AWT_IMG, ( "free Shm: %x %x %x (%dx%d)\n", img, img->xImg, img->shmiImg,
						img->width, img->height));
		destroyShmXImage( X, img, False);
	  }
	  else {
		DBG( AWT_IMG, ( "free: %x %x (%dx%d)\n", img, img->xImg, img->width, img->height));
		AWT_FREE( img->xImg->data);
		img->xImg->data = 0;
		XDestroyImage( img->xImg);
	  }
	  img->xImg = 0;
	}

	if ( img->xMask ){
	  if ( img->shmiMask ) {
		DBG( AWT_IMG, ( "free Shm mask: %x %x %x (%dx%d)\n", img, img->xMask, img->shmiMask,
						img->width, img->height));
		destroyShmXImage( X, img, True);
	  }
	  else {
		DBG( AWT_IMG, ( "free mask: %x %x (%dx%d)\n", img, img->xMask,
						img->width, img->height));
		AWT_FREE( img->xMask->data);
		img->xMask->data = 0;
		XDestroyImage( img->xMask);
	  }
	  img->xMask = 0;
	}

	if ( img->alpha ) {
	  AWT_FREE( img->alpha->buf);
	  AWT_FREE( img->alpha);
	  img->alpha = 0;
	}
	
	next = img->next;
	AWT_FREE( img);
	img = next;
  } while ( (img != 0) && (img != first) );
));
}


#if TODO  /* FOLLOWING FUNCTIONS DISABLED */

int
createShmXImage ( Toolkit* X, Image* img, int depth, int isMask )
{
  Visual  *vis = DefaultVisual( X->dsp, DefaultScreen( X->dsp));
  XShmSegmentInfo* shmi = (XShmSegmentInfo*) AWT_MALLOC( sizeof(XShmSegmentInfo));
  XImage *xim;
  unsigned int    nBytes;

  if ( isMask ) {
	xim = XShmCreateImage( X->dsp, vis, depth, XYBitmap, 0, shmi, img->width, img->height);
  }
  else {
    xim = XShmCreateImage( X->dsp, vis, depth, ZPixmap, 0, shmi, img->width, img->height);
  }

  nBytes = xim->bytes_per_line * img->height;
  shmi->shmid = shmget( IPC_PRIVATE, nBytes, (IPC_CREAT | 0777));

  /*
   * It is essential to check if shmget failed, because shared memory is usually
   * a scarce resource
   */
  if ( shmi->shmid == -1 ) {
	XShmDetach( X->dsp, shmi);
	xim->data = 0;
	XDestroyImage( xim);
	AWT_FREE( shmi);

	X->shm = SUSPEND_SHM;
	return 0;
  }

  xim->data = shmi->shmaddr = shmat( shmi->shmid, 0, 0);
  shmi->readOnly = False;
  XShmAttach( X->dsp, shmi);
  /*
   * make sure it will be freed automatically once the attachment count comes
   * down to 0 (either by explicit imgFreeImage or by process termination)
   */
  shmctl( shmi->shmid, IPC_RMID, 0);

  if ( isMask ) {
	memset( xim->data, 0xff, nBytes);
	img->shmiMask = shmi;
	img->xMask = xim;
  }
  else {
	img->shmiImg = shmi;
	img->xImg = xim;
  }

  return 1;
}


void
destroyShmXImage ( Toolkit* X, Image* img, int isMask )
{
  XShmSegmentInfo *shmi;
  XImage          *xim;

  if ( isMask ) {
	shmi = img->shmiMask;
	xim  = img->xMask;
	img->shmiMask = 0;
  }
  else {
	shmi = img->shmiImg;
	xim  = img->xImg;
	img->shmiImg = 0;
  }

  XShmDetach( X->dsp, shmi);
  xim->data = 0;
  XDestroyImage( xim);
  /* we created it as 'deleted', so we just have to detach here */
  shmdt( shmi->shmaddr);
  AWT_FREE( shmi);

  /* if we have suspended shm, give it a try again */
  if ( X->shm == SUSPEND_SHM )
	X->shm = USE_SHM;
}


void
createXImage ( Toolkit* X, Image* img )
{
UNIMPLEMENTED((
  int bitmap_pad;
  int bytes_per_line;
  int bytes_per_pix;
  unsigned int nPix;
  char *data;
  Visual *vis  = DefaultVisual( X->dsp, DefaultScreen( X->dsp));
  int    depth = DefaultDepth(  X->dsp, DefaultScreen( X->dsp));
	
  if ( depth <= 8)	      bytes_per_pix = 1;
  else if ( depth <= 16)  bytes_per_pix = 2;
  else			          bytes_per_pix = 4;

  bytes_per_line = bytes_per_pix * img->width;
  bitmap_pad = bytes_per_pix * 8;
  nPix = img->width * img->height;

  if ( (X->shm == USE_SHM) && (nPix > X->shmThreshold) && (img->alpha == 0) ) {
	if ( createShmXImage( X, img, depth, False) ){
	  DBG( AWT_IMG, ( "alloc Shm: %x %x %x (%dx%d) \n", img, img->xImg, img->shmiImg,
					  img->width, img->height));
	  return;
	}
  }

  data = AWT_CALLOC( nPix, bytes_per_pix);
  img->xImg = XCreateImage( X->dsp, vis, depth, ZPixmap, 0,
							data, img->width, img->height, bitmap_pad, bytes_per_line);
  DBG( AWT_IMG, ( "alloc: %x %x (%dx%d)\n", img, img->xImg, img->width, img->height));
));
}

void
createXMaskImage ( Toolkit* X, Image* img )
{
  int     bytes_per_line;
  unsigned int nBytes, nPix;
  char    *data;
  Visual  *vis = DefaultVisual( X->dsp, DefaultScreen( X->dsp));

  bytes_per_line = (img->width + 7) / 8;
  nPix   = img->width * img->height;
  nBytes = bytes_per_line * img->height;

  if ( (X->shm == USE_SHM) && (nPix > X->shmThreshold) ) {
	if ( createShmXImage( X, img, 1, True) ){
	  DBG( AWT_IMG, ( "alloc Shm mask: %x %x %x (%dx%d) \n", img, img->xMask, img->shmiMask,
					  img->width, img->height));
	  return;
	}
  }

  data = AWT_MALLOC( nBytes);
  memset( data, 0xff, nBytes);

  img->xMask = XCreateImage( X->dsp, vis, 1, XYBitmap, 0,
							 data, img->width, img->height, 8, bytes_per_line );
  DBG( AWT_IMG, ( "alloc mask: %x %x (%dx%d)\n", img, img->xMask, img->width, img->height));
}


void
createAlphaImage ( Toolkit* X, Image *img )
{
  int nBytes = img->width * img->height;

  img->alpha = AWT_MALLOC( sizeof( AlphaImage));

  img->alpha->width  = img->width;
  img->alpha->height = img->height;
  img->alpha->buf = AWT_MALLOC( nBytes);
  memset( img->alpha->buf, 0xff, nBytes);
}

/*
 * For images with a full alpha channel, check if we really need an alpha byte for
 * each pel, or if a mask bitmap (alpha 0x00 / 0xff) will be sufficient
 */
int
needsFullAlpha ( Toolkit* X, Image *img, double threshold )
{
  int i, j, a;
  int n = 0, max;

  if ( !img->alpha ) return 0;

  max = (img->width * img->height) * threshold;

  for ( i=0; i<img->height; i++ ) {
	for ( j=0; j<img->width; j++ ) {
	  a = GetAlpha( img->alpha, j, i);
	  if ( (a != 0) && (a != 0xff) ) {
		if ( ++n > max )
		  return 1;
	  }
	}
  }

  return 0;
}

void
countAlphas ( Image *img, int* noAlpha, int* partAlpha, int* fullAlpha )
{
  int i, j, a;

  if ( !img->alpha ) return;

  for ( i=0; i<img->height; i++ ) {
	for ( j=0; j<img->width; j++ ) {
	  a = GetAlpha( img->alpha, j, i);
	  if ( a == 0 ) {
		(*noAlpha)++;
	  }
	  else if ( a == 0xff ) {
		(*fullAlpha)++;
	  }
	  else {
		(*partAlpha)++;
	  }
	}
  }
}

/*
 * A full alpha image channel is way slower than using a mask bitmap (= 0 / 0xff alpha).
 * This function provides a simple alpha-to-mask translation
 */
void
reduceAlpha ( Toolkit* X, Image* img, int threshold )
{
  int i, j, a;

  if ( !img->alpha )
	return;

  createXMaskImage( X, img);

  for ( i=0; i<img->height; i++ ) {
	for ( j=0; j<img->width; j++ ) {
	  a = GetAlpha( img->alpha, j, i);
	  if ( a < threshold ) {
		XPutPixel( img->xImg, j, i, 0);
		XPutPixel( img->xMask, j, i, 0);
	  }
	}
  }

  AWT_FREE( img->alpha->buf);
  AWT_FREE( img->alpha);
  img->alpha = 0;
}


__inline__ int
interpolate ( int ul, int ur, int ll, int lr, double dx, double dy )
{
  double u = ul + (double)(ur - ul) * dx;
  double l = ll + (double)(lr - ll) * dx;

  return (int) (u + (l - u) * dy  + 0.5);
}

unsigned int
getScaledAlpha ( Toolkit* X, Image* img, int x, int y, double dx, double dy )
{
  int   ul, ur, ll, lr, a;
  int   xi = (dx) ? x+1 : x;
  int   yi = (dy) ? y+1 : y;

  if ( img->alpha ) {
	ul = GetAlpha( img->alpha, x, y);
	ur = GetAlpha( img->alpha, xi, y);
	ll = GetAlpha( img->alpha, x, yi);
	lr = GetAlpha( img->alpha, xi,yi);
	a = (unsigned int) interpolate( ul, ur, ll, lr, dx, dy);
	return a;
  }

  return 0xff;
}

long
getScaledPixel ( Toolkit* X, Image* img, int x, int y, double dx, double dy )
{
  unsigned long  ul, ur, ll, lr;
  int            ulR, urR, llR, lrR, ulG, urG, llG, lrG, ulB, urB, llB, lrB, r, g, b;
  int            xi = (dx) ? x+1 : x;
  int            yi = (dy) ? y+1 : y;

  if ( img->xMask ) {
	ul = XGetPixel( img->xMask, x, y);
	ur = XGetPixel( img->xMask, xi, y);
	ll = XGetPixel( img->xMask, x, yi);
	lr = XGetPixel( img->xMask, xi,yi);
	
	if ( !interpolate( ul, ur, ll, lr, dx, dy) )
	  return -1;
  }

  ul = XGetPixel( img->xImg, x, y);
  ur = XGetPixel( img->xImg, xi, y);
  ll = XGetPixel( img->xImg, x, yi);
  lr = XGetPixel( img->xImg, xi,yi);


  if ( (ul == ur) && (ll == ul) && (lr == ll) ) {
	rgbValues( X, ul, &r, &g, &b);
  }
  else {
	rgbValues( X, ul, &ulR, &ulG, &ulB);
	rgbValues( X, ur, &urR, &urG, &urB);
	rgbValues( X, ll, &llR, &llG, &llB);
	rgbValues( X, lr, &lrR, &lrG, &lrB);

	r = interpolate( ulR, urR, llR, lrR, dx, dy);
	g = interpolate( ulG, urG, llG, lrG, dx, dy);
	b = interpolate( ulB, urB, llB, lrB, dx, dy);
  }

  return pixelValue( X, (r << 16) | (g << 8) | b);
}



void
initScaledImage ( Toolkit* X, Image *tgt, Image *src,
				  int dx0, int dy0, int dx1, int dy1,
				  int sx0, int sy0, int sx1, int sy1 )
{
  double         xScale, yScale, sX, sY, sxDelta, syDelta;
  int            dx, dy, dxInc, dyInc, sx, sy;
  long           c;

  dxInc = (dx1 > dx0) ? 1 : -1;
  dyInc = (dy1 > dy0) ? 1 : -1;

  dx1 += dxInc;
  dy1 += dyInc;

  xScale = (double) (dx1 - dx0) / (double) (sx1 - sx0 +1);
  yScale = (double) (dy1 - dy0) / (double) (sy1 - sy0 +1);

  for ( dy=dy0; dy != dy1; dy += dyInc ) {
	sY = sy0 + (dy - dy0) / yScale;
	sy = (int) sY;
	syDelta = (sy < sy1) ? sY - sy : 0;

	for ( dx=dx0; dx != dx1; dx += dxInc ) {
	  sX = sx0 + (dx - dx0) / xScale;
	  sx = (int) sX;
	  sxDelta = (sx < sx1) ? sX - sx : 0;

	  if ( (c = getScaledPixel( X, src, sx, sy, sxDelta, syDelta)) != -1 ){
  		XPutPixel( tgt->xImg, dx, dy, c);
		if ( src->alpha )
		  PutAlpha( tgt->alpha, dx, dy, getScaledAlpha( X, src, sx, sy, sxDelta, syDelta));
	  }
	  else {
		XPutPixel( tgt->xMask, dx, dy, 0);
		XPutPixel( tgt->xImg, dx, dy, 0);
	  }
	}
  }
}
#endif /* TODO */


/************************************************************************************
 * exported native methods
 */


AWT_EXPORT
void*
Java_java_awt_Toolkit_imgCreateImage ( JNIEnv* env, jclass clazz, jint width, jint height )
{
	Image *img = createImage( width, height, NULL);

	DBG(AWT_IMG, ("imgCreateImage: %p (%p)\n", img, img->bitmap));
	return img;
}


AWT_EXPORT
void*
Java_java_awt_Toolkit_imgCreateScreenImage ( JNIEnv* env, jclass clazz, jint width, jint height )
{
	Image  *img = createImage( width, height, NULL);

	DBG(AWT_IMG, ("imgCreateScreenImage: %d,%d -> %p (%p) \n", width, height, img, img->bitmap ));
	return img;
}


/* generic (ImageProducer) based image construction */

typedef struct { uint8 alpha; uint8 red; uint8 green; uint8 blue; } pix32_little;
union  pix32   { pix32_little color; uint32 value; }; 

AWT_EXPORT
void
Java_java_awt_Toolkit_imgSetIdxPels ( JNIEnv* env, jclass clazz, Image * img,
									  jint x, jint y, jint w, jint h,
									  jarray clrMap, jarray idxPels, jint trans,
									  jint off, jint scan)
{
UNIMPLEMENTED((
  register int    row, col;
  unsigned long   pix;
  jint            rgb;
  jboolean        isCopy;
  jint            *clr = (*env)->GetIntArrayElements( env, clrMap, &isCopy);
  jbyte           *pel = (*env)->GetByteArrayElements( env, idxPels, &isCopy);
  unsigned char   *idx = (unsigned char*)(pel + off);
  int             maxCol = x + w;
  int             maxRow = y + h;
  unsigned char   curPel;

  if ( (trans >= 0) && !img->xMask )
	createXMaskImage( X, img);

  for ( row = y; row < maxRow; row++) {
    for ( col = x; col < maxCol; col++) {
      curPel = idx[col + row * scan];
      rgb = clr[curPel];
      pix = pixelValue( X, rgb);
      if ( trans >= 0 ) {
		if ( curPel == trans ){
		  pix = 0;
		  XPutPixel( img->xMask, col, row, 0);
		}
      }
      XPutPixel( img->xImg, col, row, pix);
    }
  }

  (*env)->ReleaseIntArrayElements( env, clrMap, clr, JNI_ABORT);
  (*env)->ReleaseByteArrayElements( env, idxPels, pel, JNI_ABORT);
));

DBG(AWT_IMG, ("imgSetIdxPels\n"));
  register int    row, col;
//  unsigned long   pix;
/*  jint            rgb; */
  pix32           rgb;
  jboolean        isCopy;
  jint            *clr = env->GetIntArrayElements( clrMap, &isCopy);
  jbyte           *pel = env->GetByteArrayElements( idxPels, &isCopy);
  unsigned char   *idx = (unsigned char*)(pel + off);
  int             maxCol = x + w;  
  int             maxRow = y + h;  
  unsigned char   curPel;

/*
  if ( (trans >= 0) && !img->xMask )
	createXMaskImage( X, img);
*/
  if (img==NULL) return;
  
  rgb_color ctrans = B_TRANSPARENT_32_BIT;
  pix32 ptrans;
  
  if (trans >= 0) {
  	/* !!! transparency, WE COULD USE MASKS */
  	ptrans.color.red  = ctrans.red;
  	ptrans.color.green= ctrans.green;
  	ptrans.color.blue = ctrans.blue;
  	ptrans.color.alpha= 0; 
  	img->trans = 0;  /* enable alpha */
  }
  
  pix32 * data = (pix32*)img->bitmap->Bits();
  
  for ( row = y; row < maxRow; row++) {
    for ( col = x; col < maxCol; col++) {
      curPel = idx[col + row * scan];
      if (trans >= 0 && curPel == trans)
      	rgb = ptrans; 
      else
      {
      	rgb.value = clr[curPel];	
      }      
      data[col + row * scan + row] = rgb;    // !!! why our scanline is one pixel shorter ?
/*
      pix = pixelValue( X, rgb);
      if ( trans >= 0 ) {
		if ( curPel == trans ){
		  pix = 0;
		  XPutPixel( img->xMask, col, row, 0);
		}
      }
      XPutPixel( img->xImg, col, row, pix);
*/
	  	  
    }
  }
  
  env->ReleaseIntArrayElements( clrMap, clr, JNI_ABORT);
  env->ReleaseByteArrayElements( idxPels, pel, JNI_ABORT);
}


AWT_EXPORT
void
Java_java_awt_Toolkit_imgSetRGBPels ( JNIEnv* env, jclass clazz, Image * img,
									  jint x, jint y, jint w, jint h,
									  jarray rgbPels, jint off, jint scan)
{

  register int    row, col;
  unsigned long   pix = 0;
  jboolean        isCopy;
  jint            *rgbs = env->GetIntArrayElements( rgbPels, &isCopy);
  jint            *rgb = rgbs + off;
  int             maxCol = x + w;
  int             maxRow = y + h;
  jint		  val;

  pix32 * data = (pix32*)img->bitmap->Bits();

  for ( row = y; row < maxRow; row++) {
    for ( col = x; col < maxCol; col++) {
	  val = rgb[col + row * scan];
	  if ( (val & 0xff000000) == 0xff000000 ) {
		pix = pixelValue( X, val);
		data[col + row * scan + row].value = pix;    // !!! why our scanline is one pixel shorter ?		
	  }
	  else {
        img->trans=0;	/* enable alpha */
		pix = pixelValue( X, val);
		data[col + row * scan + row].value = val;    // !!! why our scanline is one pixel shorter ?		
        
#if 0  // !!! TODO, SUPPORT ALPHA
		debugger("Alpha unsupported\n");
		/*
		 * No way to tell for now if this will be a reduced (on/off) or a
		 * full alpha channel. We have to be prepared for the "worst", and reduce
		 * later
		 */
		if ( !img->alpha )
		  createAlphaImage( X, img);
		PutAlpha( img->alpha, col, row, (val >> 24));
		XPutPixel( img->xImg, col, row, pix);

		/*
		if ( !img->xMask )
		  createXMaskImage( X, img);
		XPutPixel( img->xMask, col, row, 0);
		XPutPixel( img->xImg, col, row, 0);
		*/
#endif
	  }
    }
  }

DBG(AWT_IMG, ("imgSetRGBPels %p (%d,%d - %d,%d) %s\n", img, x, y, maxCol, maxRow, (img->trans>=0 ? "alpha" : "no alpha")));

  env->ReleaseIntArrayElements( rgbPels, rgbs, JNI_ABORT);
}

AWT_EXPORT
void
Java_java_awt_Toolkit_imgComplete( JNIEnv* env, jclass clazz, Image * img, jint status )
{
UNIMPLEMENTED((
  /*
   * Check for alpha channel reduction. Note that full alpha images aren't created
   * with Shm (by policy), so you might loose the Shm speed factor. This method is just
   * called for external (generic) production, since our own prod facilities usually
   * know better if and how to do alpha support
   */
  if ( img->alpha &&  !needsFullAlpha( X, img, 0.0) )
	reduceAlpha( X, img, 128);
));
DBG(AWT_IMG, ("imgComplete\n"));
}

AWT_EXPORT
void
Java_java_awt_Toolkit_imgFreeImage( JNIEnv* env, jclass clazz, Image * img)
{
DBG(AWT_IMG, ("imgFreeImage\n"));
	freeImage(X, img);
}


AWT_EXPORT
void*
Java_java_awt_Toolkit_imgCreateScaledImage ( JNIEnv* env, jclass clazz,
											 Image* img, int width, int height )
{
  Image *scaledImg = createImage( width, height, img);

DBG(AWT_IMG, ("imgCreateScaledImage %p (%d,%d) -> %p (%d,%d)\n", 
	img, img->height, img->width, scaledImg, height, width));
  return scaledImg;
}


AWT_EXPORT
void
Java_java_awt_Toolkit_imgProduceImage ( JNIEnv* env, jclass clazz, jobject producer, Image* img )
{
	
  int       i, j;
//  int       r, g, b;
//  unsigned long pix;
  jclass    prodClazz, modelClazz;
  jmethodID modelCtor,setDim,setCM,setHints,setPix,imgCompl;
  jobject   model;
  jarray	pelArray;
  jint		*pels;
  jboolean  isCopy;

DBG(AWT_IMG, ("imgProduceImage %p %d,%d\n", img, img->width, img->height));
  
  prodClazz  = env->GetObjectClass( producer);
//  jclass    modelClazz = (*env)->FindClass( env, "java/awt/image/ColorModel");
  modelClazz = env->FindClass( "kaffe/awt/JavaColorModel");
  
//  jmethodID modelCtor = (*env)->GetStaticMethodID( env, modelClazz, "getRGBdefault", "()Ljava/awt/image/ColorModel;");
  modelCtor = env->GetStaticMethodID( modelClazz, "getSingleton", "()Lkaffe/awt/JavaColorModel;");
  setDim    = env->GetMethodID( prodClazz, "setDimensions", "(II)V");
  setCM     = env->GetMethodID( prodClazz, "setColorModel", "(Ljava/awt/image/ColorModel;)V");
  setHints  = env->GetMethodID( prodClazz, "setHints", "(I)V");
  setPix    = env->GetMethodID( prodClazz, "setPixels", "(IIIILjava/awt/image/ColorModel;[III)V");
  imgCompl  = env->GetMethodID( prodClazz, "imageComplete", "(I)V");

  model     = env->CallStaticObjectMethod( modelClazz, modelCtor);

  /* for JDK compat, the pixel buffer has to be large enough to hold the *complete* image */
  pelArray  = env->NewIntArray( img->width * img->height);
  pels      = env->GetIntArrayElements( pelArray, &isCopy);

  env->CallVoidMethod( producer, setDim, img->width, img->height);
  env->CallVoidMethod( producer, setCM, model);
  env->CallVoidMethod( producer, setHints, 6); /* TOPDOWNLEFTRIGHT | COMPLETESCANLINES */

  for ( j=0; j<img->height; j++ ) {
	for ( i=0; i<img->width; i++ ) {
#if 0	// !!! TODO, MANAGE MASKS FOR REDUCED ALPHA IMAGES
	  if ( (img->xMask == 0) || XGetPixel( img->xMask, i, j) ) {
		pix = XGetPixel( img->xImg, i, j);
		rgbValues( X, pix, &r, &g, &b);
		pels[j*img->width+i] = (0xff000000 | (r << 16) | (g << 8) | b);
	  }
	  else
#endif
	  {
		pels[j*img->width+i] = 0;	/* !!! these should be set, just lacking documentation */
	  }

	}
  }
  if ( isCopy ) {
    env->ReleaseIntArrayElements( pelArray, pels, JNI_COMMIT);
  }

  env->CallVoidMethod( producer, setPix, 0, 0, img->width, img->height, model, pelArray, 0, img->width);
  env->CallVoidMethod( producer, imgCompl, 3); /* 3 = STATICIMAGEDONE */
}



/************************************************************************************
 * native format translation routines (GIF, jpeg, png, ..) using Translation Kit 
 */

/** creates an Image from a stream of data (a file or your_preferred_source_here),
    using the Translation Kit to build the bitmap 
    
    FOR GIF IMAGES WE NEED GIF TRANSLATOR INSTALLED, 
    SHOULD WE EMBED THIS CAPABILITY INTO KAFFE ? */
Image *
readImage( BPositionIO * data)
{
	Image * img = (Image *) AWT_CALLOC( 1, sizeof( Image));
	if (img==NULL) return NULL;

 	BTranslatorRoster *roster = BTranslatorRoster::Default(); 
 	BBitmapStream stream; 
	BBitmap *bmap = NULL;
	
 	if (roster->Translate(data, NULL, NULL, &stream, B_TRANSLATOR_BITMAP) >= B_OK) {
		stream.DetachBitmap(&bmap);
		if (bmap) {	
			BRect r = bmap->Bounds();

			img->trans  = -1;     /* !!! default to no alpha ? */
			img->width = (int)r.Width();  
			img->height = (int)r.Height();
		}
	}

	if (bmap==NULL) {
		DBG(AWT_IMG, ("readImage failed\n"));
		freeImage(X, img);
		return NULL;
	} 
	DBG(AWT_IMG, ("readImage: %p %d,%d\n", img, img->width, img->height));
	img->bitmap = bmap;

	return img;
}

AWT_EXPORT
void*
Java_java_awt_Toolkit_imgCreateFromFile ( JNIEnv* env, jclass clazz, jstring fileName )
{

  Image *img = 0;
  char  *fn = java2CString( env, X, fileName);

  BFile file(fn, B_READ_ONLY); 

	img = readImage(&file); 
 
 	DBG(AWT_IMG, ("imgCreateFromFile %s -> %p\n", fn, img ));
	return img;		
}

AWT_EXPORT
void*
Java_java_awt_Toolkit_imgSetFrame ( JNIEnv* env, jclass clazz, Image* img, int frameNo )
{

 	DBG(AWT_IMG, ("imgSetFrame %p %d\n", img, frameNo));

UNIMPLEMENTED((
	Image *imgCur = img;
	if ( !img->next )
		return img;

	while ( img->frame != frameNo ) {
		img = img->next;
		if ( img == imgCur ) {
			break;
		}
	}
));
	return img;
}

AWT_EXPORT
void*
Java_java_awt_Toolkit_imgCreateFromData ( JNIEnv* env, jclass clazz,
										  jarray jbuffer, jint off, jint len )
{

  Image *img = 0;
  jboolean isCopy;
  jint   length = env->GetArrayLength( jbuffer);
  jbyte  *jb = env->GetByteArrayElements( jbuffer, &isCopy);
  
  if ( (off+len) <= length)
  {
  	BMemoryIO buff(jb + off, len); 
	img = readImage(&buff);  	
  }
  
  env->ReleaseByteArrayElements( jbuffer, jb, JNI_ABORT);

  DBG(AWT_IMG, ("imgCreateFromData %p %d %d -> %p\n", jb, off, len, img));

  return img;  

}


/************************************************************************************
 * field access
 */

AWT_EXPORT
jint
Java_java_awt_Toolkit_imgGetWidth ( JNIEnv* env, jclass clazz, Image* img)
{
  return img->width;
}

AWT_EXPORT
jint
Java_java_awt_Toolkit_imgGetHeight ( JNIEnv* env, jclass clazz, Image* img)
{
  return img->height;
}

AWT_EXPORT
jboolean
Java_java_awt_Toolkit_imgIsMultiFrame ( JNIEnv* env, jclass clazz, Image* img)
{
UNIMPLEMENTED((
  return (img->next != 0);
));
  return false;
}

AWT_EXPORT
jint
Java_java_awt_Toolkit_imgGetLatency ( JNIEnv* env, jclass clazz, Image* img)
{
UNIMPLEMENTED((
  return img->latency;
));
  return 0; // !!!
}

AWT_EXPORT
void*
Java_java_awt_Toolkit_imgGetNextFrame ( JNIEnv* env, jclass clazz, Image* img )
{
UNIMPLEMENTED((
  return img->next;   /* next in the ring */
));
  return NULL; // !!!
}
