/**
 * @file wnd.cpp 
 * Native toplevel window related functions
 *
 * Copyright (c) 1999-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * Portions Copyright (c) 1999
 *      Sim Ijskes
 *
 * @author Tomaso Paoletti <tpaoletti@users.sourceforge.net>,
 *         Sim Ijskes <sim@nyx.xs4all.nl>
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.hpp"

#include <Application.h>
#include <MessageFilter.h>

extern BMessage * KWM_DESTROY;

#define WND_FRAME		0x01
#define WND_DIALOG		0x02
#define WND_WINDOW		0x04

#define WND_MAPPED		0x08
#define WND_DESTROYED	0x10

const char * RootViewName = "_bkgnd_view_";


/** The common event filter used by every native window to intercept events, see implementation in evt.cpp */
extern filter_result	windowEventFilter (BMessage * msg, BHandler ** target, BMessageFilter * msgFilter); 

/** The native widget for windows

    Strange but true, this does not directly map to a BWindow
    but to a BView which is in fact the background of a newly created BWindow.
    
    Why this hell? Because, doing this, our AWT implementation can run inside
    environments where we don't "own" windows (plugins come to mind).

	@todo use different function names for thread-safe and thread-unsafe local functions
	@todo check how to support owners in KWindow creation (see CreateWindow() and CreateDialog() )
	@todo support cursor setting by native callback (see OnSetCursor() )
	@todo implement the "root" view as a Replicant (!!)
	@todo review use of setInsets() in KWindow() 
*/
class KWindow:public KWnd {
private:
	int fType;

public:
	KWindow( BWindow * parent, char* title, int type, int x, int y, int width, int height, int clrBack) {
/*** !!!
		BRect rcl( x, y, x+width, y+height );

		// AdjustWindowRectEx( &rcl, style, FALSE, 0);

		setInsets( x-rcl.left, y-rcl.top, rcl.right-x-width, rcl.bottom-y-height, true);
		hwnd = createWindow( TEXT("window"), title, style,
					x, y, width, height,
					parent, (HMENU)0, NULL, NULL);
		defWndProc = (WNDPROC)DefWindowProc;
***/		

		BWindow *wnd;

		uint32 flags = B_NOT_ZOOMABLE ;


		/* --- create window --- */
		wnd = new BWindow( BRect(0,0,1,1), title, B_TITLED_WINDOW, flags );

		if( wnd == 0 ) {
			DBG( AWT_WND, ("KWindow %d,%d,%d,%d failed\n", x, y, width, height));
			return;
		}

		DBG( AWT_WND, ("KWindow %d,%d,%d,%d %x -> %p\n", 
				x, y, width, height, clrBack, wnd));

		fType = type;
		switch( fType ) {
			case WND_FRAME :
				wnd->SetLook( B_TITLED_WINDOW_LOOK );
				wnd->SetFeel( B_NORMAL_WINDOW_FEEL );
				break ;
			case WND_DIALOG :
				wnd->SetLook( B_MODAL_WINDOW_LOOK );
				break ;
			case WND_WINDOW :
				wnd->SetLook( B_NO_BORDER_WINDOW_LOOK );
				// we don't want a menu or drop-down list to take focus away from
				// the parent window, so we should give it a floating app feel
				
				// !!! CHEKC AGAIN
				wnd->SetFeel( B_FLOATING_APP_WINDOW_FEEL );
				break ;
		} 
	
		/* --- create the "root" view --- */
			/* every window gets a background view to draw onto */
		BView * view = new BView(wnd->Bounds(), RootViewName, B_FOLLOW_ALL | B_FULL_UPDATE_ON_RESIZE, B_WILL_DRAW);
		setView(view);
		view->SetViewColor(RGBCAST(clrBack));
		wnd->AddChild(view);

		/* --- set up object properties --- */
			
		setBounds(x, y, width, height);
UNIMPLEMENTED((
		setResizable(false);
));
		setFocus();		
		
		DBG( AWT_WND, ("KWindow properties ok\n"));

		/* --- set up event system mechanics --- */
		/* be careful that the window must be locked  */ 
		wnd->AddCommonFilter(new BMessageFilter(B_ANY_DELIVERY, B_ANY_SOURCE, windowEventFilter));
		DBG( AWT_WND, ("KWindow event filter ok\n"));

		/* --- start window's event loop --- */
			/* We unlock the Window's BLooper before its first Show() because
			   we want to let the Java app handle its own shows and hides
			   ! This avoids the potential deadlock in graInitGraphics()
			   that happens right after Window creation ! */
		wnd->BLooper::Run();
		
		DBG( AWT_WND, ("KWindow looper started\n"));
	}

		
	void KWindow::setBounds(int x, int y, int w, int h) {
		BRect frame;

		/* for menus and drop-down windows, make coordinates relative to
		parent window; otherwise just use coordinates passed to function */

		BWindow * win = getWindow();
		if (fType == WND_WINDOW) {
			frame = win->Frame();
			frame.left += x;
			frame.top += y;
			frame.right = frame.left + w;
			frame.bottom = frame.top + h; 
			win->MoveTo(frame.LeftTop());
		} else {
			win->MoveTo(x,y);
		}
		win->ResizeTo(w,h);
	}
	
		/** A POTENTIAL SOURCE OF DEAD-LOCKS
		
		    when we hide the window, its thread is suspended,
		    so every other Java thread plotting on this window
		    is blocked, and some other resource may never be released 
		*/
	void KWindow::setVisible( jboolean b) {
		BWindow * w = getWindow();
		DBG(AWT_WND, ("KWindow setVisible %d (%d)\n", b, w->IsMinimized() ));
		if (b) {
			w->Show();			
		} else {			
			// !!! Hide();
		}
	}

	void KWindow::setInsets( int xl, int yt, int xr,int yb, bool notify) {
		jclass frCls;
		
		jmethodID setDecoInsets;
		
		insets.left = xl;
		insets.right = xr;
		insets.bottom = yb;
		insets.top = yt;
		
		if ( notify) {
			frCls = JniEnv->FindClass( "java/awt/Frame");
			setDecoInsets = JniEnv->GetStaticMethodID( frCls, "setDecoInsets", "(IIIII)V");
			
			X->frameInsets.left = xl;
			X->frameInsets.right = xr;
			X->frameInsets.bottom = yb;
			X->frameInsets.top = yt;
			
			JniEnv->CallStaticVoidMethod( frCls, setDecoInsets, yt, xl, yb, xr, -1);
		}
	}
	
	void KWindow::setResizable(bool sw) {
		/** !!! SET ZOOMABLE FLAG ?		**/
		
		uint32 flags = getWindow()->Flags();
		flags &= ~B_NOT_RESIZABLE ;
		if( !sw ) {
			flags = B_NOT_RESIZABLE ;
		}
		getWindow()->SetFlags( flags );
	}

	void KWindow::setTitle(char * str) {
		getWindow()->SetTitle(str);
	}

	bool KWindow::OnPaint( int x, int y, int width, int height){
				
		//map to frame extends
		width  += (int)insets.left + (int)insets.right;
		height += (int)insets.bottom + (int)insets.top;
		
		//if (erase) // !!!
/*
		if (lock()) {
			fillRect( x,y,width,height));
			unlock();
		}
*/
		X->jEvt = JniEnv->CallStaticObjectMethod( PaintEvent, getPaintEvent,
					srcIdx, PAINT, //ps.fErase ? UPDATE : PAINT,
					x, y, width, height );			
		return true;
	}
	
	bool KWindow::OnActivate( int flags, bool min, BWindow * handle ) {
		//handled by java		
		return false;
	}

	
	bool KWindow::OnCommand( int cmd) {
UNIMPLEMENTED((
		HMENU hmenu = GetMenu( hwnd);
#if defined(UNDER_CE)
		if ( !hmenu && hbar )
			hmenu = CommandBar_GetMenu( hbar, 0);
#endif
		if ( hmenu )
			processMenuAction( hmenu, cmd);
));
		return false;
	}
	
	
	bool KWindow::OnPosChanged( int x, int y, int width, int height) {
		X->jEvt = JniEnv->CallStaticObjectMethod( ComponentEvent, getComponentEvent,
			srcIdx, COMPONENT_RESIZED,
			x, y, width, height) ;		
		return true;
	}
	
	
	bool KWindow::OnClose() {
		X->jEvt = JniEnv->CallStaticObjectMethod( WindowEvent, getWindowEvent,
			srcIdx, WINDOW_CLOSING);		
		return true;
	}
		
	bool KWindow::OnSetCursor() {
		//getWindow()->SetCursor( cursor);				
		return TRUE;
	}
	
	bool KWindow::updateInsets() {
	/*
	* called by Menu assign to detect appropriate
	* client insets
		*/
#if 0	/* no need to adjust window rect */
		RECT rcl, rclc;
		
		GetClientRect( hwnd, &rclc);
		
		rcl.left = rclc.left;
		rcl.top  = rclc.top;
		rcl.right = rclc.right;
		rcl.bottom = rclc.bottom;

#if !defined(UNDER_CE)
		AdjustWindowRectEx( &rclc, GetWindowLong( hwnd, GWL_STYLE), (GetMenu( hwnd) != NULL), 0);
#else
		AdjustWindowRectEx( &rclc, GetWindowLong( hwnd, GWL_STYLE), 0, 0);
		if ( hbar )
			rclc.top += CommandBar_Height( hbar);
#endif

		setInsets( rcl.left-rclc.left,
				rcl.top-rclc.top,
				rclc.right-rcl.right,
				rclc.bottom-rcl.bottom, false );
		
#endif
		setInsets( 0,0,0,0, false);		// !!! check this
		return TRUE;
	}
	
};


extern "C" {
	
	void __cdecl
		Java_java_awt_Toolkit_wndSetTitle ( JNIEnv* env, jclass clazz, KWindow* wnd, jstring jTitle )
	{
		char * title;
		if ( jTitle) {
			title = java2CString( JniEnv, X, jTitle);
		} else {
			title="";
		}
		wnd->setTitle(title);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_wndSetResizable ( JNIEnv* env, jclass clazz, KWindow* wnd, jboolean isResizable,
		int x, int y, int width, int height )
	{
		DBG( AWT_WND, ("setResizable: %p, %d, %d,%d,%d,%d\n", wnd, isResizable, x,y,width,height));
		wnd->setResizable( isResizable );
	}
	
	void* __cdecl
		Java_java_awt_Toolkit_wndCreateFrame ( JNIEnv* env, jclass clazz, jstring jTitle,
		jint x, jint y, jint width, jint height,
		jint jCursor, jint clrBack, jboolean isResizable )
	{		
		DBG( AWT_WND, ("createFrame( %p, %d,%d,%d,%d %x)\n", jTitle,x,y,width,height,clrBack));
		
		KWindow* wnd;
		char* title;

#if !defined(D_LOOP_MODE)
/* !!!
		if ( !X->dspTid )
			X->dspTid = GetCurrentThreadId();
*/
#endif

		title = java2CString( env, X, jTitle);
		wnd = new KWindow( NULL, title, WND_FRAME, x, y, width, height, clrBack);
		
		DBG( AWT_WND, (" -> %p\n", wnd));
		
		if ( !isResizable ) {
			Java_java_awt_Toolkit_wndSetResizable( env, clazz,
				wnd, FALSE, x, y, width, height);			
		}
		
		return wnd;
		
	}
	
	
	void* __cdecl
		Java_java_awt_Toolkit_wndCreateWindow ( JNIEnv* env, jclass clazz, KWindow* owner,
		jint x, jint y, jint width, jint height,
		jint jCursor, jint clrBack )
	{
		KWindow* wnd;
		
		DBG( AWT_WND, ("createWindow( %p, %d,%d,%d,%d %x)\n", owner,x,y,width,height, clrBack));
		
		wnd = new KWindow( NULL /* !!! UNSUPPORTED owner ? UNHAND(owner) : NULL */, "", WND_WINDOW, x, y, width, height, clrBack);
		
		DBG( AWT_WND, (" -> %p\n", wnd));
		
		return wnd;
	}
	
	void* __cdecl		
		Java_java_awt_Toolkit_wndCreateDialog ( JNIEnv* env, jclass clazz, KWindow* owner, jstring jTitle,
		jint x, jint y, jint width, jint height,
		jint jCursor, jint clrBack, jboolean isResizable )
	{
		DBG( AWT_WND, ("createDialog( %p,%p, %d,%d,%d,%d %x)\n", owner,jTitle,x,y,width,height,clrBack));
		
		KWindow* wnd;
		TCHAR* title;
				
		title = java2CString( env, X, jTitle);
		wnd = new KWindow( NULL /* !!! unsupported UNHAND(owner)*/, title, WND_DIALOG, x, y, width, height, clrBack);
		
		DBG( AWT_WND, (" -> %p\n", wnd));
				
		if ( !isResizable ) {			
			Java_java_awt_Toolkit_wndSetResizable( env, clazz,				
				wnd, FALSE, x, y, width, height);						
		}
				
		return wnd;
		
	}
	
	void __cdecl
		Java_java_awt_Toolkit_wndDestroyWindow ( JNIEnv* env, jclass clazz, KWindow* wnd )
	{
		DBG( AWT_WND, ("destroy window: %p\n", wnd));
		wnd->getWindow()->PostMessage( KWM_DESTROY);
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_wndRequestFocus ( JNIEnv* env, jclass clazz, KWindow* wnd )
	{
		DBG( AWT_WND, ("request focus: %p\n", wnd));
		wnd->setFocus();
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_wndSetFrameInsets ( JNIEnv* env, jclass clazz,
		jint top, jint left, jint bottom, jint right )
	{
		X->frameInsets.top  = top;
		X->frameInsets.left = left;
		X->frameInsets.bottom = bottom;
		X->frameInsets.right = right;
		X->frameInsets.guess = 1;
	}
	
	void __cdecl
		Java_java_awt_Toolkit_wndSetDialogInsets ( JNIEnv* env, jclass clazz,
		jint top, jint left, jint bottom, jint right )
	{
		X->dialogInsets.top  = top;
		X->dialogInsets.left = left;
		X->dialogInsets.bottom = bottom;
		X->dialogInsets.right = right;
		X->dialogInsets.guess = 1;
	}
	
	void __cdecl
		Java_java_awt_Toolkit_wndSetBounds ( JNIEnv* env, jclass clazz, KWindow* wnd,
		jint x, jint y, jint width, jint height,
		jboolean isResizable )
	{		
		DBG( AWT_WND, ("setBounds: %p %d,%d,%d,%d %d\n", wnd, x, y, width, height, isResizable));
				
		if ( width < 0 )  width = 1;		
		if ( height < 0 ) height = 1;
		
		wnd->setBounds(x, y, width, height);

		if ( !isResizable ) {
			Java_java_awt_Toolkit_wndSetResizable( env, clazz,
				wnd, FALSE, x, y, width, height);
		}
	}
	
	
	
	void __cdecl
		Java_java_awt_Toolkit_wndRepaint ( JNIEnv* env, jclass clazz, KWindow* wnd,
		jint x, jint y, jint width, jint height )
	{
		DBG( AWT_WND, ("wndRepaint: %p %d,%d,%d,%d\n", wnd, x, y, width, height));
		
		BView * bgnd = wnd->getView();
		if (bgnd) {
			bgnd->Invalidate(BRect(x, y, x+width, y+height));
		}
	}
	
	
	
	void __cdecl
		Java_java_awt_Toolkit_wndSetIcon ( JNIEnv* env, jclass clazz, KWindow* wnd, void* img )
	{	
	}
	
	
	void __cdecl
		Java_java_awt_Toolkit_wndSetVisible ( JNIEnv* env, jclass clazz, KWindow* wnd, jboolean showIt )
	{
		DBG( AWT_WND, ("setVisible: %p %d\n", wnd, showIt));
		wnd->setVisible( showIt);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_wndToBack ( JNIEnv* env, jclass clazz, KWindow* wnd )
	{
		DBG( AWT_WND, ("toBack: %p\n", wnd));
		wnd->getWindow()->Activate(false);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_wndToFront ( JNIEnv* env, jclass clazz, KWindow* wnd )
	{
		DBG( AWT_WND, ("toFront: %p\n", wnd));
		wnd->getWindow()->Activate(true);
	}
	
	void __cdecl
		Java_java_awt_Toolkit_wndSetCursor ( JNIEnv* env, jclass clazz, KWindow* wnd, jint jCursor )
	{
		DBG( AWT_WND, ("setCursor: %p, %d\n", wnd, jCursor));
		be_app->SetCursor( getCursor( jCursor));
	}
	
}
